import dataclasses
import inspect
import logging
import os
import sys
import threading
from types import FrameType
import warnings
import copy
from abc import ABC, abstractmethod
from collections import defaultdict
from operator import getitem
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Generic,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import react_ipywidgets as react
import reacton.core
from solara.util import equals_extra

import solara
import solara.settings
from solara import _using_solara_server
from solara.util import nullcontext

T = TypeVar("T")
TS = TypeVar("TS")
S = TypeVar("S")  # used for state
logger = logging.getLogger("solara.toestand")

_DEBUG = False


class ThreadLocal(threading.local):
    reactive_used: Optional[Set["ValueBase"]] = None
    reactive_watch: Optional[Callable[["ValueBase"], None]] = None


thread_local = ThreadLocal()


# these hooks should go into react-ipywidgets
def use_sync_external_store(subscribe: Callable[[Callable[[], None]], Callable[[], None]], get_snapshot: Callable[[], Any]):
    _, set_counter = react.use_state(0)

    def force_update():
        set_counter(lambda x: x + 1)

    state = get_snapshot()
    prev_state = react.use_ref(state)

    def update_state():
        prev_state.current = state

    react.use_effect(update_state)

    def on_store_change(_ignore_new_state=None):
        new_state = get_snapshot()
        if not equals_extra(new_state, prev_state.current):
            prev_state.current = new_state
            force_update()

    react.use_effect(lambda: subscribe(on_store_change), [])
    return state


def use_sync_external_store_with_selector(subscribe, get_snapshot: Callable[[], Any], selector):
    return use_sync_external_store(subscribe, lambda: selector(get_snapshot()))


def merge_state(d1: S, **kwargs) -> S:
    if dataclasses.is_dataclass(d1):
        return dataclasses.replace(d1, **kwargs)  # type: ignore
    if "pydantic" in sys.modules and isinstance(d1, sys.modules["pydantic"].BaseModel):
        module = sys.modules["pydantic"]
        version_major = int(module.__version__.split(".")[0])
        if version_major >= 2:
            return d1.model_copy(update=kwargs)
        else:
            return d1.copy(update=kwargs)
    return cast(S, {**cast(dict, d1), **kwargs})


class ValueBase(Generic[T]):
    def __init__(self, merge: Callable = merge_state, equals=equals_extra):
        self.merge = merge
        self.equals = equals
        self.listeners: Dict[str, Set[Tuple[Callable[[T], None], Optional[ContextManager]]]] = defaultdict(set)
        self.listeners2: Dict[str, Set[Tuple[Callable[[T, T], None], Optional[ContextManager]]]] = defaultdict(set)

    # make sure all boolean operations give type errors
    if not solara.settings.main.allow_reactive_boolean:

        def __bool__(self):
            raise TypeError("Reactive vars are not allowed in boolean expressions, did you mean to use .value?")

        def __eq__(self, other):
            raise TypeError(f"'==' not supported between a Reactive and {other.__class__.__name__}, did you mean to use .value?")

        def __ne__(self, other):
            raise TypeError(f"'!=' not supported between a Reactive and {other.__class__.__name__}, did you mean to use .value?")

        # If we explicitly define __eq__, we need to explicitly define __hash__ as well
        # Otherwise our class is marked unhashable
        __hash__ = object.__hash__

    def __lt__(self, other):
        raise TypeError(f"'<' not supported between a Reactive and {other.__class__.__name__}, did you mean to use .value?")

    def __le__(self, other):
        raise TypeError(f"'<=' not supported between a Reactive and {other.__class__.__name__}, did you mean to use .value?")

    def __gt__(self, other):
        raise TypeError(f"'>' not supported between a Reactive and {other.__class__.__name__}, did you mean to use .value?")

    def __ge__(self, other):
        raise TypeError(f"'>=' not supported between a Reactive and {other.__class__.__name__}, did you mean to use .value?")

    def __len__(self):
        raise TypeError("'len(...)' is not supported for a Reactive, did you mean to use .value?")

    @property
    def lock(self):
        raise NotImplementedError

    @property
    def value(self) -> T:
        return self.get()

    @value.setter
    def value(self, value: T):
        self.set(value)

    def set(self, value: T):
        raise NotImplementedError

    def peek(self) -> T:
        raise NotImplementedError

    def get(self) -> T:
        raise NotImplementedError

    def _get_scope_key(self):
        raise NotImplementedError

    def subscribe(self, listener: Callable[[T], None], scope: Optional[ContextManager] = None):
        if scope is not None:
            warnings.warn("scope argument should not be used, it was only for internal use")
        del scope
        scope_id = self._get_scope_key()
        rc = reacton.core.get_render_context(required=False)
        if _using_solara_server():
            import solara.server.kernel_context

            kernel = solara.server.kernel_context.get_current_context() if solara.server.kernel_context.has_current_context() else nullcontext()
        else:
            kernel = nullcontext()
        context = Context(rc, kernel)

        self.listeners[scope_id].add((listener, context))

        def cleanup():
            self.listeners[scope_id].remove((listener, context))

        return cleanup

    def subscribe_change(self, listener: Callable[[T, T], None], scope: Optional[ContextManager] = None):
        if scope is not None:
            warnings.warn("scope argument should not be used, it was only for internal use")
        del scope
        scope_id = self._get_scope_key()
        rc = reacton.core.get_render_context(required=False)
        if _using_solara_server():
            import solara.server.kernel_context

            kernel = solara.server.kernel_context.get_current_context() if solara.server.kernel_context.has_current_context() else nullcontext()
        else:
            kernel = nullcontext()
        context = Context(rc, kernel)
        self.listeners2[scope_id].add((listener, context))

        def cleanup():
            self.listeners2[scope_id].remove((listener, context))

        return cleanup

    def fire(self, new: T, old: T):
        logger.info("value change from %s to %s, will fire events", old, new)
        scope_id = self._get_scope_key()
        contexts = set()
        for listener, context in self.listeners[scope_id].copy():
            contexts.add(context)
        for listener2, context in self.listeners2[scope_id].copy():
            contexts.add(context)
        if contexts:
            for context in contexts:
                with context or nullcontext():
                    for listener, context_listener in self.listeners[scope_id].copy():
                        if context == context_listener:
                            listener(new)
                    for listener2, context_listener in self.listeners2[scope_id].copy():
                        if context == context_listener:
                            listener2(new, old)

    def update(self, _f=None, **kwargs):
        if _f is not None:
            assert not kwargs
            with self.lock:
                kwargs = _f(self.get())
        with self.lock:
            # important to have this part thread-safe
            new = self.merge(self.get(), **kwargs)
            self.set(new)

    def use_value(self) -> T:
        # .use with the default argument doesn't give good type inference
        return self.use()

    def use(self, selector: Callable[[T], TS] = lambda x: x) -> TS:  # type: ignore
        return selector(self.value)

    def use_state(self) -> Tuple[T, Callable[[T], None]]:
        setter = self.set
        value = self.use()  # type: ignore
        return value, setter

    @property
    def fields(self) -> T:
        # we lie about the return type, but in combination with
        # setter we can make type safe setters (see docs/tests)
        return cast(T, Fields(self))

    def setter(self, field: TS) -> Callable[[TS], None]:
        _field = cast(FieldBase, field)

        def setter(new_value: TS):
            _field.set(new_value)

        return cast(Callable[[TS], None], setter)

    def _check_mutation(self):
        pass


# the default store for now, stores in a global dict, or when in a solara
# context, in the solara user context


class KernelStore(ValueBase[S], ABC):
    _global_dict: Dict[str, S] = {}  # outside of solara context, this is used
    # we keep a counter per type, so the storage keys we generate are deterministic
    _type_counter: Dict[Any, int] = defaultdict(int)
    scope_lock = threading.RLock()

    def __init__(self, key: str, equals: Callable[[Any, Any], bool] = equals_extra):
        super().__init__(equals=equals)
        self.storage_key = key
        self._global_dict = {}
        # since a set can trigger events, which can trigger new updates, we need a recursive lock
        self._lock = threading.RLock()
        self.local = threading.local()

    @property
    def lock(self):
        return self._lock

    def _get_scope_key(self):
        scope_dict, scope_id = self._get_dict()
        return scope_id

    def _get_dict(self):
        scope_dict = self._global_dict
        scope_id = "global"
        if _using_solara_server():
            import solara.server.kernel_context

            try:
                context = solara.server.kernel_context.get_current_context()
            except RuntimeError:  # noqa
                pass  # do we need to be more strict?
            else:
                scope_dict = cast(Dict[str, S], context.user_dicts)
                scope_id = context.id
        return cast(Dict[str, S], scope_dict), scope_id

    def peek(self):
        return self.get()

    def get(self):
        scope_dict, scope_id = self._get_dict()
        if self.storage_key not in scope_dict:
            with self.scope_lock:
                if self.storage_key not in scope_dict:
                    # we assume immutable, so don't make a copy
                    scope_dict[self.storage_key] = self.initial_value()
        return scope_dict[self.storage_key]

    def clear(self):
        scope_dict, scope_id = self._get_dict()
        if self.storage_key in scope_dict:
            del scope_dict[self.storage_key]

    def set(self, value: S):
        scope_dict, scope_id = self._get_dict()
        old = self.get()
        if self.equals(old, value):
            return
        scope_dict[self.storage_key] = value

        if _DEBUG:
            import traceback

            traceback.print_stack(limit=17, file=sys.stdout)

            print("change old", old)  # noqa
            print("change new", value)  # noqa

        self.fire(value, old)

    @abstractmethod
    def initial_value(self) -> S:
        pass

    def _check_mutation(self):
        pass


def _is_internal_module(file_name: str):
    file_name_parts = file_name.split(os.sep)
    if len(file_name_parts) < 2:
        return False
    return (
        file_name_parts[-2:] == ["solara", "toestand.py"]
        or file_name_parts[-2:] == ["solara", "reactive.py"]
        or file_name_parts[-2:] == ["solara", "tasks.py"]
        or file_name_parts[-2:] == ["solara", "_stores.py"]
        or file_name_parts[-3:] == ["solara", "hooks", "use_reactive.py"]
        or file_name_parts[-2:] == ["reacton", "core.py"]
        # If we use SomeClass[K](...) we go via the typing module, so we need to skip that as well
        or (file_name_parts[-2].startswith("python") and file_name_parts[-1] == "typing.py")
    )


def _find_outside_solara_frame() -> Optional[FrameType]:
    # the module where the call stack origined from
    current_frame: Optional[FrameType] = None
    module_frame = None

    # _getframe is not guaranteed to exist in all Python implementations,
    # but is much faster than the inspect module
    if hasattr(sys, "_getframe"):
        current_frame = sys._getframe(1)
    else:
        current_frame = inspect.currentframe()

    while current_frame is not None:
        file_name = current_frame.f_code.co_filename
        # Skip most common cases, i.e. toestand.py, reactive.py, use_reactive.py, Reacton's core.py, and the typing module
        if not _is_internal_module(file_name):
            module_frame = current_frame
            break
        current_frame = current_frame.f_back

    return module_frame


class KernelStoreValue(KernelStore[S]):
    default_value: S
    _traceback: Optional[inspect.Traceback]
    _default_value_copy: Optional[S]

    def __init__(self, default_value: S, key=None, equals: Callable[[Any, Any], bool] = equals_extra, unwrap=lambda x: x):
        self.default_value = default_value
        self._unwrap = unwrap
        self.equals = equals
        self._mutation_detection = solara.settings.storage.mutation_detection
        if self._mutation_detection:
            frame = _find_outside_solara_frame()
            if frame is not None:
                self._traceback = inspect.getframeinfo(frame)
            else:
                self._traceback = None
            self._default_value_copy = copy.deepcopy(default_value)
            if not self.equals(self._unwrap(self.default_value), self._unwrap(self._default_value_copy)):
                msg = """The equals function for this reactive value returned False when comparing a deepcopy to itself.

This reactive variable will not be able to detect mutations correctly, and is therefore disabled.

To avoid this warning, and to ensure that mutation detection works correctly, please provide a better equals function to the reactive variable.
A good choice for dataframes and numpy arrays might be solara.util.equals_pickle, which will also attempt to compare the pickled values of the objects.

Example:
df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
reactive_df = solara.reactive(df, equals=solara.util.equals_pickle)
"""
                tb = self._traceback
                if tb:
                    if tb.code_context:
                        code = tb.code_context[0]
                    else:
                        code = "<No code context available>"
                    msg += f"This warning was triggered from:\n{tb.filename}:{tb.lineno}\n{code}"
                warnings.warn(msg)
                self._mutation_detection = False
        cls = type(default_value)
        if key is None:
            with KernelStoreValue.scope_lock:
                index = self._type_counter[cls]
                self._type_counter[cls] += 1
            key = cls.__module__ + ":" + cls.__name__ + ":" + str(index)
        super().__init__(key=key, equals=equals)

    def initial_value(self) -> S:
        self._check_mutation()
        return self.default_value

    def _check_mutation(self):
        if not self._mutation_detection:
            return
        initial = self._unwrap(self._default_value_copy)
        current = self._unwrap(self.default_value)
        if not self.equals(initial, current):
            tb = self._traceback
            if tb:
                if tb.code_context:
                    code = tb.code_context[0].strip()
                else:
                    code = "No code context available"
                msg = f"Reactive variable was initialized at {tb.filename}:{tb.lineno} with {initial!r}, but was mutated to {current!r}.\n{code}"
            else:
                msg = f"Reactive variable was initialized with a value of {initial!r}, but was mutated to {current!r} (unable to report the location in the source code)."
            raise ValueError(msg)


def _create_key_callable(f: Callable):
    try:
        prefix = f.__qualname__
    except Exception:
        prefix = repr(f)
    with KernelStore.scope_lock:
        index = KernelStore._type_counter[prefix]
        KernelStore._type_counter[prefix] += 1
    try:
        key = f.__module__ + ":" + prefix + ":" + str(index)
    except Exception:
        key = prefix + ":" + str(index)
    return key


class KernelStoreFactory(KernelStore[S]):
    def __init__(self, factory: Callable[[], S], key=None, equals: Callable[[Any, Any], bool] = equals_extra):
        self.factory = factory
        key = key or _create_key_callable(factory)
        super().__init__(key=key, equals=equals)

    def initial_value(self) -> S:
        return self.factory()


def mutation_detection_storage(default_value: S, key=None, equals=None) -> ValueBase[S]:
    from solara.util import equals_pickle as default_equals
    from ._stores import MutateDetectorStore, StoreValue, _PublicValueNotSet, _SetValueNotSet

    kernel_store = KernelStoreValue[StoreValue[S]](
        StoreValue[S](private=default_value, public=_PublicValueNotSet(), get_traceback=None, set_value=_SetValueNotSet(), set_traceback=None),
        key=key,
        unwrap=lambda x: x.private,
    )
    return MutateDetectorStore[S](kernel_store, equals=equals or default_equals)


def default_storage(default_value: S, key=None, equals=None) -> ValueBase[S]:
    # in solara v2 we will also do this when mutation_detection is None
    # and we do not run on production mode
    if solara.settings.storage.mutation_detection is True:
        return mutation_detection_storage(default_value, key=key, equals=equals)
    else:
        return KernelStoreValue[S](default_value, key=key, equals=equals or equals_extra)


def _call_storage_factory(default_value: S, key=None, equals=None) -> ValueBase[S]:
    factory = solara.settings.storage.get_factory()
    return factory(default_value, key=key, equals=equals)


class Reactive(ValueBase[S]):
    _storage: ValueBase[S]

    def __init__(self, default_value: Union[S, ValueBase[S]], key=None, equals=None):
        super().__init__()
        if not isinstance(default_value, ValueBase):
            self._storage = _call_storage_factory(default_value, key=key, equals=equals)
        else:
            self._storage = default_value
        self.__post__init__()
        self._name = None
        self._owner = None

    def __set_name__(self, owner, name):
        self._name = name
        self._owner = owner

    def __repr__(self):
        value = self.peek()
        if self._name:
            return f"<Reactive {self._owner.__name__}.{self._name} value={value!r} id={hex(id(self))}>"
        else:
            return f"<Reactive value={value!r} id={hex(id(self))}>"

    def __str__(self):
        if self._name:
            return f"{self._owner.__name__}.{self._name}={self.value!r}"
        else:
            return f"{self.value!r}"

    @property
    def lock(self):
        return self._storage.lock

    def __post__init__(self):
        pass

    def update(self, *args, **kwargs):
        self._storage.update(*args, **kwargs)

    def set(self, value: S):
        if value is self:
            raise ValueError("Can't set a reactive to itself")
        self._storage.set(value)

    def get(self, add_watch=None) -> S:
        if add_watch is not None:
            warnings.warn("add_watch is deprecated, use .peek()", DeprecationWarning)
        if thread_local.reactive_used is not None:
            thread_local.reactive_used.add(self)
        if thread_local.reactive_watch is not None:
            thread_local.reactive_watch(self)
        # peek to avoid parents also adding themselves to the reactive_used set
        return self._storage.peek()

    def peek(self) -> S:
        """Return the value without automatically subscribing to listeners."""
        return self._storage.peek()

    def subscribe(self, listener: Callable[[S], None], scope: Optional[ContextManager] = None):
        return self._storage.subscribe(listener, scope=scope)

    def subscribe_change(self, listener: Callable[[S, S], None], scope: Optional[ContextManager] = None):
        return self._storage.subscribe_change(listener, scope=scope)

    def computed(self, f: Callable[[S], T]) -> "Computed[T]":
        def func():
            return f(self.get())

        return Computed(func, key=f.__qualname__)


class Singleton(Reactive[S]):
    _storage: KernelStore[S]

    def __init__(self, factory: Callable[[], S], key=None):
        import solara.lifecycle

        super().__init__(KernelStoreFactory(factory, key=key))

        # reset on kernel restart (e.g. hot reload)
        def reset():
            def cleanup():
                self._storage.clear()

            return cleanup

        solara.lifecycle.on_kernel_start(reset)

    def __set__(self, obj, value):
        raise AttributeError("Can't set a singleton")


class Computed(Reactive[S]):
    _storage: KernelStore[S]

    def __init__(self, f: Callable[[], S], key=None):
        import solara.lifecycle

        self.f = f

        def on_change(*ignore):
            with self._auto_subscriber.value:
                self.set(f())

        import functools

        self._auto_subscriber = Singleton(functools.wraps(AutoSubscribeContextManager)(lambda: AutoSubscribeContextManager(on_change)))

        @functools.wraps(f)
        def factory():
            v = self._auto_subscriber.value
            with v:
                return f()

        super().__init__(KernelStoreFactory(factory, key=key))

        # reset on kernel restart (e.g. hot reload)
        def reset():
            def cleanup():
                self._storage.clear()

            return cleanup

        solara.lifecycle.on_kernel_start(reset)

    def __repr__(self):
        value = super().__repr__()
        return "<Computed" + value[len("<Reactive") : -1]


@overload
def computed(
    f: None,
    *,
    key: Optional[str] = ...,
) -> Callable[[Callable[[], T]], Reactive[T]]: ...


@overload
def computed(
    f: Callable[[], T],
    *,
    key: Optional[str] = ...,
) -> Reactive[T]: ...


def computed(
    f: Union[None, Callable[[], T]],
    *,
    key: Optional[str] = None,
) -> Union[Callable[[Callable[[], T]], Reactive[T]], Reactive[T]]:
    """Creates a reactive variable that is set to the return value of the function.

    The value will be updated when any of the reactive variables used in the function
    change.

    ## Example

    ```solara
    import solara
    import solara.lab


    a = solara.reactive(1)
    b = solara.reactive(2)

    @solara.lab.computed
    def total():
        return a.value + b.value

    def reset():
        a.value = 1
        b.value = 2

    @solara.component
    def Page():
        print(a, b, total)
        solara.IntSlider("a", value=a)
        solara.IntSlider("b", value=b)
        solara.Text(f"a + b = {total.value}")
        solara.Button("reset", on_click=reset)
    ```

    """

    def wrapper(f: Callable[[], T]):
        return Computed(f, key=key)

    if f is None:
        return wrapper
    else:
        return wrapper(f)


class ReactiveField(Reactive[T]):
    def __init__(self, field: "FieldBase", equals: Callable[[Any, Any], bool] = equals_extra):
        # super().__init__()  # type: ignore
        # We skip the Reactive constructor, because we do not need it, but we do
        # want to be an instanceof for use in use_reactive
        ValueBase.__init__(self, equals=equals)
        self._field = field
        field = field
        while not isinstance(field, ValueBase):
            field = field._parent
        self._root = field
        assert isinstance(self._root, ValueBase)

    def __str__(self):
        return str(self._field)

    def __repr__(self):
        return f"<Reactive field {self._field}>"

    @property
    def lock(self):
        return self._root.lock

    def subscribe(self, listener: Callable[[T], None], scope: Optional[ContextManager] = None):
        def on_change(new, old):
            try:
                new_value = self._field.get(new)
            except IndexError:
                return  # the current design choice to silently drop the update message
            except KeyError:
                return  # same
            old_value = self._field.get(old)
            if not self.equals(new_value, old_value):
                listener(new_value)

        return self._root.subscribe_change(on_change, scope=scope)

    def subscribe_change(self, listener: Callable[[T, T], None], scope: Optional[ContextManager] = None):
        def on_change(new, old):
            try:
                new_value = self._field.get(new)
            except IndexError:
                return  # see subscribe
            except KeyError:
                return  # see subscribe
            old_value = self._field.get(old)
            if not self.equals(new_value, old_value):
                listener(new_value, old_value)

        return self._root.subscribe_change(on_change, scope=scope)

    def get(self, add_watch=None) -> T:
        if add_watch is not None:
            warnings.warn("add_watch is deprecated, use .peek()", DeprecationWarning)
        if thread_local.reactive_used is not None:
            thread_local.reactive_used.add(self)
        if thread_local.reactive_watch is not None:
            thread_local.reactive_watch(self)
        # peek to avoid parents also adding themselves to the reactive_used set
        return self._field.peek()

    def peek(self) -> T:
        return self._field.peek()

    def set(self, value: T):
        self._field.set(value)

    def update(self, *args, **kwargs):
        ValueBase.update(cast(ValueBase, self), *args, **kwargs)


def Ref(field: T) -> Reactive[T]:
    _field = cast(FieldBase, field)
    return cast(Reactive[T], ReactiveField[T](_field))


class FieldBase:
    _parent: Any

    def __getattr__(self, key):
        if key in ["_parent", "set", "_lock"] or key.startswith("__"):
            return self.__dict__[key]
        return FieldAttr(self, key)

    def __getitem__(self, key):
        return FieldItem(self, key)

    def get(self, obj=None):
        raise NotImplementedError

    def set(self, value):
        raise NotImplementedError


class Fields(FieldBase):
    def __init__(self, state: ValueBase):
        self._parent = state
        self._lock = state.lock

    def get(self, obj=None):
        # we are at the root, so override the object
        # so we can get the 'old' value
        if obj is not None:
            return obj
        return self._parent.get()

    def peek(self, obj=None):
        # we are at the root, so override the object
        # so we can get the 'old' value
        if obj is not None:
            return obj
        return self._parent.peek()

    def set(self, value):
        self._parent.set(value)

    def __repr__(self):
        return repr(self._parent)


class FieldAttr(FieldBase):
    def __init__(self, parent, key: str):
        self._parent = parent
        self.key = key
        self._lock = parent._lock

    def get(self, obj=None):
        obj = self._parent.get(obj)
        return getattr(obj, self.key)

    def peek(self, obj=None):
        obj = self._parent.peek(obj)
        return getattr(obj, self.key)

    def set(self, value):
        with self._lock:
            parent_value = self._parent.peek()
            if isinstance(self.key, str):
                parent_value = merge_state(parent_value, **{self.key: value})
                self._parent.set(parent_value)
            else:
                raise TypeError(f"Type of key {self.key!r} is not supported")

    def __str__(self):
        return f".{self.key}"

    def __repr__(self):
        return f"<Field {self._parent}{self}>"


class FieldItem(FieldBase):
    def __init__(self, parent, key: str):
        self._parent = parent
        self.key = key
        self._lock = parent._lock

    def get(self, obj=None):
        obj = self._parent.get(obj)
        return getitem(obj, self.key)

    def peek(self, obj=None):
        obj = self._parent.peek(obj)
        return getitem(obj, self.key)

    def set(self, value):
        with self._lock:
            parent_value = self._parent.peek()
            if isinstance(self.key, int) and isinstance(parent_value, (list, tuple)):
                parent_type = type(parent_value)
                parent_value = parent_value.copy()  # type: ignore
                parent_value[self.key] = value
                self._parent.set(parent_type(parent_value))
            else:
                parent_value = merge_state(parent_value, **{self.key: value})
                self._parent.set(parent_value)


class AutoSubscribeContextManagerBase:
    # a render loop might trigger a new render loop of a differtent render context
    # so we want to save, and restore the current reactive_used
    reactive_used: Optional[Set[ValueBase]] = None
    subscribed: Dict[ValueBase, Callable]
    subscribed_previous_run: Dict[ValueBase, Callable]
    on_change: Callable[[], None]
    previous_reactive_watch: Optional[Callable[["ValueBase"], None]] = None

    def __init__(self):
        self.subscribed = {}
        self.subscribed_previous_run = {}
        self.on_change = lambda: None

    def unsubscribe_previous(self):
        removed = set(self.subscribed_previous_run or set()) - set(self.subscribed)
        if removed:
            for reactive in removed:
                unsubscribe = self.subscribed_previous_run[reactive]
                unsubscribe()
                del self.subscribed_previous_run[reactive]

    def add(self, reactive: ValueBase):
        relevant_reactive = reactive
        if isinstance(reactive, ValueSubField):
            root = reactive._root
            if root in self.subscribed or root in self.subscribed_previous_run:
                # we already subscribed to this reactive's root
                return
            else:
                # we are subscribing to this reactive's root
                pass

        # TODO: we could see if we are the root of any of the subscribed fields,
        # and remove that field.
        if relevant_reactive not in self.subscribed:
            if relevant_reactive not in self.subscribed_previous_run:
                unsubscribe = relevant_reactive.subscribe_change(lambda *args: self.on_change())
                self.subscribed[relevant_reactive] = unsubscribe
            else:
                self.subscribed[relevant_reactive] = self.subscribed_previous_run[relevant_reactive]

    def unsubscribe_all(self):
        for reactive in self.subscribed:
            unsubscribe = self.subscribed[reactive]
            unsubscribe()

    def __enter__(self):
        self.subscribed = {}
        self.reactive_used_before = thread_local.reactive_used
        self.previous_reactive_watch = thread_local.reactive_watch
        thread_local.reactive_watch = self.add
        self.reactive_used = thread_local.reactive_used = set()

    def __exit__(self, exc_type, exc_val, exc_tb):
        thread_local.reactive_used = self.reactive_used_before
        thread_local.reactive_watch = self.previous_reactive_watch
        self.unsubscribe_previous()
        self.subscribed_previous_run = self.subscribed.copy()


class Context:
    def __init__(self, render_context, kernel_context):
        # combine the render context *and* the kernel context into one context
        self.render_context = render_context
        self.kernel_context = kernel_context

    def __enter__(self):
        if self.render_context is not None:
            self.render_context.__enter__()
        self.kernel_context.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.render_context is not None:
            # this will trigger a render
            res1 = self.render_context.__exit__(exc_type, exc_val, exc_tb)
        else:
            res1 = None

        # pop the current context from the stack
        res2 = self.kernel_context.__exit__(exc_type, exc_val, exc_tb)
        return res1 or res2

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Context):
            return False
        return self.render_context == value.render_context and self.kernel_context == value.kernel_context

    def __hash__(self) -> int:
        return hash(id(self.render_context)) ^ hash(id(self.kernel_context))

    def __repr__(self) -> str:
        return f"Context(render_context={self.render_context}, kernel_context={self.kernel_context})"


class AutoSubscribeContextManagerReacton(AutoSubscribeContextManagerBase):
    def __init__(self, element: solara.Element):
        self.element = element
        super().__init__()

    def __enter__(self):
        _, set_counter = solara.use_state(0, key="auto_subscribe_force_update_counter")

        def force_update():
            # can we do just x+1 to collapse multiple updates into one?
            set_counter(lambda x: x + 1)

        super().__enter__()
        self.on_change = force_update

        def on_close():
            def cleanup():
                self.unsubscribe_all()

            return cleanup

        solara.use_effect(on_close, [])


class AutoSubscribeContextManager(AutoSubscribeContextManagerBase):
    def __init__(self, on_change: Callable[[], None]):
        super().__init__()
        self.on_change = on_change


# alias for compatibility
State = Reactive
ValueSubField = ReactiveField

auto_subscribe_context_manager = AutoSubscribeContextManagerReacton
reacton.core._component_context_manager_classes.append(auto_subscribe_context_manager)



"""
# ipyleaflet
Map visualization using [ipyleaflet](https://ipyleaflet.readthedocs.io/), a ipywidgets wrapper for [leaflet.js](https://leafletjs.com/)
"""

import ipyleaflet

import solara

zoom = solara.reactive(5)
center = solara.reactive((53.2305799, 6.5323552))
bounds = solara.reactive(None)


@solara.component
def Page():
    # Isolation is required to prevent the map from overlapping navigation (when screen width < 960px)
    with solara.Column(style={"width": "100%", "height": "500px", "isolation": "isolate"}):
        # solara components support reactive variables
        solara.SliderInt(label="Zoom level", value=zoom, min=1, max=20)
        # using 3rd party widget library require wiring up the events manually
        # using zoom.value and zoom.set
        ipyleaflet.Map.element(  # type: ignore
            zoom=zoom.value,
            on_zoom=zoom.set,
            center=center.value,
            on_center=center.set,
            on_bounds=bounds.set,
            scroll_wheel_zoom=True,
        )
        solara.Text(f"Zoom: {zoom.value}")
        solara.Text(f"Center: {center.value}")
        solara.Text(f"Bounds: {bounds.value}")



import solara

github_url = solara.util.github_url(__file__)
# some app state that outlives a single page
app_state = solara.reactive(0)


@solara.component
def SharedComponent():
    with solara.Card("Shown on each page", style={"max-width": "500px"}, margin=0, classes=["my-2"]):
        solara.Markdown(
            f"""
            This component will be used on each page.

            It uses the `app_state` [reactive variable](https://solara.dev/documentation/api/utilities/reactive)
            so that the state outlives each page


            app_state: {app_state.value}
            """
        )
        solara.Button(label="Increment app_state", icon_name="mdi-plus", on_click=lambda: app_state.set(app_state.value + 1), outlined=True)


@solara.component
def Page():
    with solara.Sidebar():
        solara.Markdown("This is the sidebar at the home page!")
    with solara.Card("Home"):
        solara.Markdown("This is the home page")
        solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)

        SharedComponent()


@solara.component
def Layout(children):
    return solara.AppLayout(children=children)



import json
import logging
import os
from pathlib import Path

import tornado.web
from jupyter_server.base.handlers import JupyterHandler
import solara.server.server as server

from solara.server.utils import path_is_child_of
import solara

logger = logging.getLogger("solara.server.jupyter.solara")


@solara.component
def Page():
    solara.Error("Hi, you should not see this, we only support ipypopout for now")


class SolaraHandler(JupyterHandler):
    async def get(self, path=None):
        try:
            # base url ends with /
            base_url = self.settings["base_url"]
            # root_path's do not end with /
            jupyter_root_path = ""
            if base_url and base_url.endswith("/"):
                jupyter_root_path = base_url[:-1]
            root_path = f"{jupyter_root_path}/solara"
            content = server.read_root(path="", root_path=root_path, jupyter_root_path=jupyter_root_path)
        except Exception as e:
            logger.exception(e)
            raise tornado.web.HTTPError(500)

        if content is None:
            raise tornado.web.HTTPError(404)

        self.set_header("Content-Type", "text/html")
        self.write(content)


# similar to voila
class MultiStaticFileHandler(tornado.web.StaticFileHandler):
    """A static file handler that 'merges' a list of directories

    If initialized like this::

        application = web.Application([
            (r"/content/(.*)", web.MultiStaticFileHandler, {"paths": ["/var/1", "/var/2"]}),
        ])

    A file will be looked up in /var/1 first, then in /var/2.

    """

    def initialize(self, paths, default_filename=None):  # type: ignore
        self.roots = paths
        super().initialize(path=paths[0], default_filename=default_filename)

    def get_absolute_path(self, root: str, path: str) -> str:  # type: ignore
        # find the first absolute path that exists
        self.root = self.roots[0]
        abspath = os.path.abspath(os.path.join(root, path))
        for root in self.roots[1:]:
            abspath = os.path.abspath(os.path.join(root, path))
            if os.path.exists(abspath):
                self.root = root  # make sure all the other methods in the base class know how to find the file
                break

        # tornado probably already does a version of this, to make sure it behaves as the rest of the solara
        # server, we do it again
        if not path_is_child_of(Path(abspath), Path(self.root)):
            raise PermissionError(f"Trying to read from outside of cache directory: {abspath} is not a subdir of {self.root}")

        return abspath


class Assets(MultiStaticFileHandler):
    def initialize(self):  # type: ignore
        super().initialize(server.asset_directories())
        logging.error("Using %r as assets directories", self.roots)


class ReadyZ(JupyterHandler):
    def get(self):
        json_data, status = server.readyz()
        json_response = json.dumps(json_data)
        self.set_header("Content-Type", "application/json")
        self.set_status(status)
        self.write(json_response)



import solara


@solara.component
def Page():
    volume, set_volume = solara.use_state(5)
    with solara.VBox() as main:
        solara.Markdown("# Choose your volume")
        solara.IntSlider("Volume", value=volume, on_value=set_volume, min=0, max=11)
        if volume == 11:
            solara.Markdown(f"Yeah! {volume} is the best volume!")
    return main



import threading
from abc import abstractmethod
from typing import Any, Callable, Dict, List, MutableMapping
from warnings import warn

ObserverCallback = Callable[[Any, Any], None]


class MutableMappingBase(MutableMapping):
    @abstractmethod
    def _get_dict(self) -> MutableMapping:
        pass

    def __delitem__(self, key) -> None:
        self._get_dict().__delitem__(key)

    def __getitem__(self, key):
        return self._get_dict().__getitem__(key)

    def __iter__(self):
        return self._get_dict().__iter__()

    def __len__(self):
        return self._get_dict().__len__()

    def __setitem__(self, key, value):
        self._get_dict().__setitem__(key, value)


class ObservableMutableMapping(MutableMappingBase):
    # list of observers for each key
    observers: Dict[Any, List[ObserverCallback]]

    def __init__(self) -> None:
        super().__init__()
        self.observers = {}
        self.lock = threading.Lock()

    def subscribe_key(self, key, callback: ObserverCallback):
        if key not in self.observers:
            with self.lock:
                if key not in self.observers:
                    self.observers[key] = []
        self.observers[key].append(callback)
        if self.observers[key].count(callback) > 1:
            warn("Callback already subscribed")

        def unsubscribe():
            self.observers[key].remove(callback)

        return unsubscribe

    def trigger_key(self, key):
        for callback in self.observers[key]:
            callback(key, self[key])



import inspect

import solara
from solara.alias import rw
from solara.components import Markdown


@solara.component
def Sample(code, component):
    locals = globals().copy()
    exec(code, locals)
    c = locals[component]
    with rw.VBox() as main:
        Markdown(
            f"""
```python
{code}
```
"""
        )
        c()
    return main


@solara.component
def IncludeComponent(component, pre="", highlight=[], **kwargs):
    code = inspect.getsource(component.f)
    with rw.VBox(layout={"padding": "20px", "max_width": "1024px", "border": "1px #333 solid"}) as main:
        Markdown(
            f"""
```python
{pre}{code}
```
""",
            highlight=highlight,
        )
        component(**kwargs)
    return main



import solara
from solara.website.components.contact import Contact


@solara.component
def Page():
    with solara.Column(gap="40px", align="stretch", style={"width": "100%"}):
        solara.Markdown("""
# Pricing

The main Solara package is free for both personal and commercial use. There are additional components and functionality available in the `solara-enterprise` package, which is free for personal use, but requires a license for commercial use.
        """)
        solara.Markdown("""
## Enterprise License

As stated before, the `solara-enterprise` package is free for personal use, but requires a license for commercial use. The license includes the following benefits:

- Access to all `solara-enterprise` components and functionality
- Priority support from the Solara team, including bug fixes and questions
- Potential for custom components and functionality as agreed separately
- Trainings and workshops for your team
- Regular code reviews and/or design consultation

All enterprise licenses are custom and are priced based on the needs of the organization. Please contact us below for more information.
        """)

        Contact(
            title="Get In Touch",
            subtitle="Please fill out the form below and we will get back to you as soon as possible.",
            email_subject="Enterprise License Inquiry",
        )



import ipyvue
import traitlets

import solara


class SpinnerSolaraWidget(ipyvue.VueTemplate):
    template_file = (__file__, "spinner-solara.vue")

    size = traitlets.Unicode("64px").tag(sync=True)
    color_back = traitlets.Unicode("#FFCF64").tag(sync=True)
    color_front = traitlets.Unicode("#FF8C3E").tag(sync=True)


@solara.component
def SpinnerSolara(size="64px", color_back="#FFCF64", color_front="#FF8C3E"):
    """Spinner component with the Solara logo to indicate the app is busy.

    ## Examples
    ### Basic example

    ```solara
    import solara

    @solara.component
    def Page():
        solara.SpinnerSolara(size="100px")
    ```

    ## Changing the colors
    ```solara
    import solara

    @solara.component
    def Page():
        solara.SpinnerSolara(size="100px", color_back="Grey", color_front="Lime")
    ```


    ## Arguments
     * `size`: Size of the spinner.
     * `color_back`: Color of the spinner in the background.
     * `color_front`: Color of the spinner in the foreground.
    """
    return SpinnerSolaraWidget.element(size=size, color_back=color_back, color_front=color_front)



import reacton.ipyvuetify as rv
import solara

from ..data import articles


@solara.component
def ArticleCard(name):
    article = articles[name]
    with rv.Card(max_width="400px") as main:
        with solara.Link(f"/article/{name}"):
            rv.Img(height="250", src=article.image_url)
        rv.CardTitle(children=[article.title])
        with rv.CardText():
            solara.Markdown(article.description)
            with solara.Link(f"/article/{name}"):
                solara.Button("Read article", text=True, icon_name="mdi-book-open")
    return main


@solara.component
def Overview():
    with solara.ColumnsResponsive(12) as main:
        with solara.Card("Company articles"):
            with solara.ColumnsResponsive(12, small=6, large=4):
                for name in articles:
                    ArticleCard(name)
    return main



import contextlib
import inspect
import json
import logging
import os
import shlex
import subprocess
import sys
import textwrap
import threading
import typing
import tempfile
import urllib.parse
import uuid
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Union, cast

import ipywidgets as widgets
import pytest
import requests
from IPython.display import display

import solara.server.app
import solara.server.kernel_context
import solara.server.server
import solara.server.settings
from solara.server import reload
from solara.server.starlette import ServerStarlette
from solara.server.threaded import ServerBase

if typing.TYPE_CHECKING:
    import playwright.sync_api


logger = logging.getLogger("solara.pytest_plugin")

TEST_PORT_START = int(os.environ.get("PORT", "18765")) + 100  # do not interfere with the solara integration tests
TEST_HOST = solara.server.settings.main.host
TIMEOUT = float(os.environ.get("SOLARA_PW_TIMEOUT", "18"))
PYTEST_IPYWIDGETS_SOLARA_APP_WAIT_TIMEOUT = int(os.environ.get("PYTEST_IPYWIDGETS_SOLARA_APP_WAIT_TIMEOUT", "10"))
runners = os.environ.get("SOLARA_TEST_RUNNERS", "solara,voila,jupyter_lab,jupyter_notebook").split(",")


@pytest.fixture(scope="session")
def solara_server(pytestconfig: Any, request):
    port = pytestconfig.getoption("--solara-port")
    host = pytestconfig.getoption("--solara-host")
    webserver = ServerStarlette(port, host)

    try:
        webserver.serve_threaded()
        webserver.wait_until_serving()
        yield webserver
    finally:
        webserver.stop_serving()


try:
    from pytest_playwright.pytest_playwright import connect_options  # noqa

    has_pytest_070 = True
except ImportError:
    has_pytest_070 = False


if has_pytest_070:
    # pytest-playwright >= 0.7.0
    from pytest_playwright.pytest_playwright import Browser, CreateContextCallback, BrowserContext, Playwright, ArtifactsRecorder, _truncate_file_name
    from slugify import slugify

    @pytest.fixture(scope="session")
    def output_path_session(pytestconfig: Any, request: pytest.FixtureRequest) -> str:
        output_dir = Path(pytestconfig.getoption("--output")).absolute()
        return os.path.join(output_dir, _truncate_file_name(slugify(request.node.nodeid)))

    @pytest.fixture(scope="session")
    def _artifacts_recorder_session(
        request: pytest.FixtureRequest,
        output_path_session: str,
        playwright: Playwright,
        pytestconfig: Any,
        _pw_artifacts_folder: tempfile.TemporaryDirectory,
    ) -> Generator["ArtifactsRecorder", None, None]:
        artifacts_recorder = ArtifactsRecorder(pytestconfig, request, output_path_session, playwright, _pw_artifacts_folder)
        yield artifacts_recorder
        # If request.node is missing rep_call, then some error happened during execution
        # that prevented teardown, but should still be counted as a failure
        failed = request.node.rep_call.failed if hasattr(request.node, "rep_call") else True
        artifacts_recorder.did_finish_test(failed)

    @pytest.fixture(scope="session")
    def new_context_session(
        browser: Browser,
        browser_context_args: Dict,
        _artifacts_recorder_session: "ArtifactsRecorder",
        request: pytest.FixtureRequest,
    ) -> Generator[CreateContextCallback, None, None]:
        browser_context_args = browser_context_args.copy()
        context_args_marker = next(request.node.iter_markers("browser_context_args"), None)
        additional_context_args = context_args_marker.kwargs if context_args_marker else {}
        browser_context_args.update(additional_context_args)
        contexts: List[BrowserContext] = []

        def _new_context(**kwargs: Any) -> BrowserContext:
            context = browser.new_context(**browser_context_args, **kwargs)
            original_close = context.close

            def _close_wrapper(*args: Any, **kwargs: Any) -> None:
                contexts.remove(context)
                _artifacts_recorder_session.on_will_close_browser_context(context)
                original_close(*args, **kwargs)

            context.close = _close_wrapper
            contexts.append(context)
            _artifacts_recorder_session.on_did_create_browser_context(context)
            return context

        yield cast(CreateContextCallback, _new_context)
        for context in contexts.copy():
            context.close()

    @pytest.fixture(scope="session")
    def context_session(new_context_session):
        return new_context_session()
else:

    @pytest.fixture(scope="session")
    def context_session(
        browser: "playwright.sync_api.Browser",
        browser_context_args: Dict,
        pytestconfig: Any,
        request: pytest.FixtureRequest,
    ) -> Generator["playwright.sync_api.BrowserContext", None, None]:
        from playwright.sync_api import Error, Page
        from pytest_playwright.pytest_playwright import _build_artifact_test_folder
        from slugify import slugify  # type: ignore

        pages: List[Page] = []
        context = browser.new_context(**browser_context_args)
        context.on("page", lambda page: pages.append(page))

        tracing_option = pytestconfig.getoption("--tracing")
        capture_trace = tracing_option in ["on", "retain-on-failure"]
        if capture_trace:
            context.tracing.start(
                title=slugify(request.node.nodeid),
                screenshots=True,
                snapshots=True,
                sources=True,
            )

        # decrease timeout from 30s to 10s
        context.set_default_timeout(TIMEOUT * 1000)
        yield context

        # If request.node is missing rep_call, then some error happened during execution
        # that prevented teardown, but should still be counted as a failure
        failed = request.node.rep_call.failed if hasattr(request.node, "rep_call") else True

        if capture_trace:
            retain_trace = tracing_option == "on" or (failed and tracing_option == "retain-on-failure")
            if retain_trace:
                trace_path = _build_artifact_test_folder(pytestconfig, request, "trace.zip")
                context.tracing.stop(path=trace_path)
            else:
                context.tracing.stop()

        screenshot_option = pytestconfig.getoption("--screenshot")
        capture_screenshot = screenshot_option == "on" or (failed and screenshot_option == "only-on-failure")
        if capture_screenshot:
            for index, page in enumerate(pages):
                human_readable_status = "failed" if failed else "finished"
                screenshot_path = _build_artifact_test_folder(pytestconfig, request, f"test-{human_readable_status}-{index + 1}.png")
                try:
                    page.screenshot(timeout=5000, path=screenshot_path)
                except Error:
                    pass

        context.close()

        video_option = pytestconfig.getoption("--video")
        preserve_video = video_option == "on" or (failed and video_option == "retain-on-failure")
        if preserve_video:
            for page in pages:
                video = page.video
                if not video:
                    continue
                try:
                    video_path = video.path()
                    file_name = os.path.basename(video_path)
                    video.save_as(path=_build_artifact_test_folder(pytestconfig, request, file_name))
                except Error:
                    # Silent catch empty videos.
                    pass


# page fixture that keeps open all the time, is faster
@pytest.fixture(scope="session")
def page_session(context_session: "playwright.sync_api.BrowserContext"):
    page = context_session.new_page()

    def log(msg):
        print("BROWSER LOG:", msg.text)  # noqa
        logger.debug("PAGE LOG: %s", msg.text)

    page.on("console", log)
    yield page
    page.close()


@pytest.fixture()
def solara_app(solara_server):
    used_app = None

    @contextlib.contextmanager
    def run(app: Union[solara.server.app.AppScript, str], init=True):
        nonlocal used_app
        if "__default__" in solara.server.app.apps:
            solara.server.app.apps["__default__"].close()
        if isinstance(app, str):
            app = solara.server.app.AppScript(app)
            if init:
                app.init()
        used_app = app
        solara.server.app.apps["__default__"] = app
        try:
            yield
        finally:
            if app.type == solara.server.app.AppType.MODULE:
                if app.name in sys.modules and app.name.startswith("tests.integration.testapp"):
                    del sys.modules[app.name]
                if app.name in reload.reloader.watched_modules:
                    reload.reloader.watched_modules.remove(app.name)

    yield run
    if used_app:
        used_app.close()


run_events: Dict[str, threading.Event] = {}
used_contexts: Dict[str, solara.server.kernel_context.VirtualKernelContext] = {}


@solara.component
def SyncWrapper():
    global run_calls
    router = solara.use_router()
    values = urllib.parse.parse_qs(router.search, keep_blank_values=True)
    id = values.get("id", [None])[0]  # type: ignore
    if id is None:
        solara.Error("No id found in url")
    else:
        import reacton.ipywidgets as w

        used_contexts[id] = solara.server.kernel_context.get_current_context()
        run_events[id].set()

        return w.VBox(children=[w.HTML(value="Test in solara"), w.VBox()])


@contextlib.contextmanager
def _solara_test(solara_server, solara_app, page_session: "playwright.sync_api.Page", require_vuetify_warmup: bool):
    with solara_app("solara.test.pytest_plugin:SyncWrapper"):
        id = str(uuid.uuid4())
        run_events[id] = run_event = threading.Event()
        page_session.goto(solara_server.base_url + f"?id={id}")
        try:
            assert run_event.wait(PYTEST_IPYWIDGETS_SOLARA_APP_WAIT_TIMEOUT)
            context = used_contexts[id]
            with context:
                test_output_warmup = widgets.Output()
                test_output = widgets.Output()
                try:
                    page_session.locator("text=Test in solara").wait_for()
                    assert context.container
                    context.container.children[0].children[1].children[1].children = [test_output_warmup]  # type: ignore
                    with test_output_warmup:
                        page_session.add_style_tag(
                            content="""
                            .solara-content-main {
                                animation-duration: 0s !important;
                            }
                        """
                        )
                        if require_vuetify_warmup:
                            warmup()
                            button = page_session.locator(".solara-warmup-widget")
                            button.wait_for()
                            page_session.evaluate("document.fonts.ready")
                            button.click()
                            button.wait_for(state="detached")
                            page_session.evaluate("document.fonts.ready")
                    context.container.children[0].children[1].children[1].children = [test_output]  # type: ignore
                    with test_output:
                        yield
                finally:
                    test_output.close()
                    test_output_warmup.close()
        finally:
            del run_events[id]
            page_session.goto("about:blank")
            if id in used_contexts:
                # handle when run_event.wait(10) fails
                del used_contexts[id]
                assert context.closed_event.wait(10)


@pytest.fixture()
def solara_test(solara_server, solara_app, page_session: "playwright.sync_api.Page", pytestconfig: Any):
    require_vuetify_warmup = pytestconfig.getoption("solara_vuetify_warmup")
    with _solara_test(solara_server, solara_app, page_session, require_vuetify_warmup):
        yield


class ServerVoila(ServerBase):
    popen = None

    def __init__(self, notebook_path, port: int, host: str = "localhost", **kwargs):
        self.notebook_path = notebook_path
        super().__init__(port, host)

    def has_started(self):
        try:
            return requests.get(self.base_url).status_code // 100 in [2, 3]
        except requests.exceptions.ConnectionError:
            return False

    def signal_stop(self):
        if self.popen is None:
            return
        self.popen.terminate()
        self.popen.kill()

    def serve(self):
        if self.has_started():
            raise RuntimeError(f"Jupyter server already running, use lsof -i :{self.port} to find the process and kill it")
        cmd = (
            "voila --no-browser --VoilaTest.log_level=DEBUG --Voila.port_retries=0 --VoilaExecutor.timeout=240"
            f' --Voila.port={self.port} --show_tracebacks=True "{self.notebook_path}" --enable_nbextensions=True'
        )
        logger.info(f"Starting Voila server at {self.base_url} with command {cmd}")
        if sys.platform == "win32":
            args = cmd
        else:
            args = shlex.split(cmd)
        self.popen = subprocess.Popen(args, shell=False, stdout=sys.stdout, stderr=sys.stderr, stdin=None)
        self.started.set()


class ServerJupyter(ServerBase):
    popen = None

    def __init__(self, notebook_path, port: int, host: str = "localhost", **kwargs):
        self.notebook_path = notebook_path
        super().__init__(port, host)

    def has_started(self):
        try:
            return requests.get(self.base_url).status_code // 100 in [2, 3]
        except requests.exceptions.ConnectionError:
            return False

    def signal_stop(self):
        if self.popen is None:
            return
        self.popen.terminate()
        self.popen.kill()

    def serve(self):
        if self.has_started():
            raise RuntimeError(f"Jupyter server already running, use lsof -i :{self.port} to find the process and kill it")
        cmd = f'jupyter lab --port={self.port} --no-browser --ServerApp.token="" --port-retries=0 "{self.notebook_path}"'
        logger.info(f"Starting Jupyter (lab) server at {self.base_url} with command {cmd}")
        if sys.platform == "win32":
            args = cmd
        else:
            args = shlex.split(cmd)
        self.popen = subprocess.Popen(args, shell=False, stdout=sys.stdout, stderr=sys.stderr, stdin=None)
        self.started.set()


@pytest.fixture(scope="session")
def voila_server(pytestconfig: Any, notebook_path):
    if "voila" not in runners:
        yield None
        return
    port = pytestconfig.getoption("--voila-port")
    host = pytestconfig.getoption("--solara-host")
    write_notebook(["print('hello')"], notebook_path)
    server = ServerVoila(notebook_path, port, host)
    try:
        server.serve_threaded()
        server.wait_until_serving()
        yield server
    finally:
        server.stop_serving()


@pytest.fixture(scope="session")
def jupyter_server(pytestconfig: Any, notebook_path):
    if "jupyter_lab" not in runners and "jupyter_notebook" not in runners:
        yield None
        return
    port = pytestconfig.getoption("--jupyter-port")
    host = pytestconfig.getoption("--solara-host")
    write_notebook(["print('hello')"], notebook_path)
    server = ServerJupyter(notebook_path, port, host)
    try:
        server.serve_threaded()
        server.wait_until_serving()
        yield server
    finally:
        server.stop_serving()


def code_from_function(f) -> str:
    lines = inspect.getsourcelines(f)[0]
    lines = lines[1:]
    return textwrap.dedent("".join(lines))


def write_notebook(codes: List[str], path: str):
    notebook = {
        "cells": [{"cell_type": "code", "execution_count": None, "id": "df77670d", "metadata": {}, "outputs": [], "source": [code]} for code in codes],
        "metadata": {
            "kernelspec": {"display_name": "Python 3 (ipykernel)", "language": "python", "name": "python3"},
            "language_info": {
                "codemirror_mode": {"name": "ipython", "version": 3},
                "file_extension": ".py",
                "mimetype": "text/x-python",
                "name": "python",
                "nbconvert_exporter": "python",
                "pygments_lexer": "ipython3",
                "version": "3.9.16",
            },
        },
        "nbformat": 4,
        "nbformat_minor": 5,
    }

    with open(path, "w") as file:
        json.dump(notebook, file)


@pytest.fixture(scope="session")
def notebook_path(tmp_path_factory):
    path = tmp_path_factory.mktemp("notebooks") / "notebook.ipynb"
    yield str(path)


def warmup():
    import ipyvuetify as v

    # give it a scope so we do not collide with the user's code
    def scoped():
        button = v.Btn(children=["Warmup js/css/fonts", v.Icon(children=["mdi-check"])], class_="solara-warmup-widget")
        container = v.Container(
            children=[button],
            class_="ma-2 snapshot-container",
        )

        def remove_button(*ignore):
            container.children = []
            container.style_ = "display: none"

        button.on_event("click", remove_button)
        return container

    display(scoped())


def create_runner_voila(voila_server, notebook_path, page_session: "playwright.sync_api.Page", require_vuetify_warmup: bool):
    count = 0
    base_url = voila_server.base_url

    def run(f: Callable, locals={}):
        nonlocal count
        path = Path(f.__code__.co_filename)
        cwd = str(path.parent)
        code_setup = f"""
import os
os.chdir({cwd!r})
        \n"""
        for name, value in locals.items():
            code_setup += f"{name} = {value!r}\n"
        if require_vuetify_warmup:
            write_notebook([code_setup, code_from_function(warmup), code_from_function(f)], notebook_path)
        else:
            write_notebook([code_setup, code_from_function(f)], notebook_path)
        page_session.goto(base_url + f"?v={count}")
        if require_vuetify_warmup:
            button = page_session.locator(".solara-warmup-widget")
            button.wait_for()
            page_session.evaluate("document.fonts.ready")
            button.click()
            button.wait_for(state="detached")
            page_session.evaluate("document.fonts.ready")
        count += 1

    return run


def create_runner_jupyter_lab(jupyter_server, notebook_path, page_session: "playwright.sync_api.Page", require_vuetify_warmup: bool):
    count = 0
    base_url = jupyter_server.base_url

    def run(f: Callable, locals={}):
        nonlocal count
        path = Path(f.__code__.co_filename)
        cwd = str(path.parent)
        code_setup = f"""
import os
os.chdir({cwd!r})
        \n"""
        if require_vuetify_warmup:
            code_setup += """
import ipyvuetify as v;
display(v.Btn(children=['Warmup js/css/fonts', v.Icon(children=["mdi-check"])], class_="solara-warmup-widget"))
        \n"""
        for name, value in locals.items():
            code_setup += f"{name} = {value!r}\n"

        write_notebook([code_setup, code_from_function(f)], notebook_path)
        page_session.goto(base_url + f"/lab/workspaces/solara-test/tree/notebook.ipynb?reset&v={count}")
        page_session.locator('css=[data-command="runmenu:run"]').wait_for()
        # close the file browser tab, it does not give a consistent width each time
        # which leads to fractional pixel x, which causes the screenshot to be different
        # by 1 pixel
        # first make sure the page is loaded
        page_session.locator('css=[data-id="filebrowser"]').wait_for()
        page_session.locator('css=[data-command="filebrowser:create-main-launcher"]').wait_for()
        # close
        page_session.locator('css=[data-id="filebrowser"]').click()
        # make sure it is closed
        page_session.locator('css=[data-command="filebrowser:create-main-launcher"]').wait_for(state="hidden")

        page_session.locator('button:has-text("No Kernel")').wait_for(state="detached")
        page_session.locator('css=[data-status="idle"]').wait_for()
        page_session.locator('css=[data-command="runmenu:run"]').click()
        if require_vuetify_warmup:
            page_session.locator(".solara-warmup-widget").wait_for()
            page_session.evaluate("document.fonts.ready")
        page_session.locator('css=[data-command="runmenu:run"]').click()
        count += 1

    return run


def create_runner_jupyter_notebook(jupyter_server, notebook_path, page_session: "playwright.sync_api.Page", require_vuetify_warmup: bool):
    count = 0
    base_url = jupyter_server.base_url

    def run(f: Callable, locals={}):
        nonlocal count
        path = Path(f.__code__.co_filename)
        cwd = str(path.parent)
        code_setup = f"""
import os
os.chdir({cwd!r})
"""
        if require_vuetify_warmup:
            code_setup += """
import ipyvuetify as v;
display(v.Btn(children=['Warmup js/css/fonts', v.Icon(children=["mdi-check"])], class_="solara-warmup-widget"))
        \n"""
        for name, value in locals.items():
            code_setup += f"{name} = {value!r}\n"
        write_notebook([code_setup, code_from_function(f)], notebook_path)
        page_session.goto(base_url + f"/notebooks/notebook.ipynb?v={count}")
        page_session.locator(".prompt_container >> nth=0").wait_for()
        page_session.locator("text=Kernel starting, please wait...").wait_for(state="detached")
        page_session.locator("Kernel Ready").wait_for(state="detached")
        page_session.locator('css=[data-jupyter-action="jupyter-notebook:run-cell-and-select-next"]').click()
        if require_vuetify_warmup:
            page_session.locator(".solara-warmup-widget").wait_for()
            page_session.evaluate("document.fonts.ready")
        page_session.locator('css=[data-jupyter-action="jupyter-notebook:run-cell-and-select-next"]').click()
        count += 1

    return run


@contextlib.contextmanager
def create_runner_solara(solara_server, solara_app, page_session: "playwright.sync_api.Page", require_vuetify_warmup: bool):
    count = 0

    def run(f: Callable, locals={}):
        nonlocal count
        from IPython.display import clear_output

        path = Path(f.__code__.co_filename)
        cwd = str(path.parent)
        current_dir = os.getcwd()
        os.chdir(cwd)
        import sys

        sys.path.append(cwd)
        try:
            clear_output()
            f()
        finally:
            os.chdir(current_dir)
            sys.path.remove(cwd)
        count += 1

    with _solara_test(solara_server, solara_app, page_session, require_vuetify_warmup):
        yield run


@pytest.fixture(params=runners)
def ipywidgets_runner(
    solara_server,
    solara_app,
    voila_server,
    jupyter_server,
    notebook_path,
    page_session: "playwright.sync_api.Page",
    request,
    pytestconfig: Any,
):
    runner = request.param
    require_vuetify_warmup = pytestconfig.getoption("solara_vuetify_warmup")
    if runner == "solara":
        with create_runner_solara(solara_server, solara_app, page_session, require_vuetify_warmup) as runner:
            yield runner
    elif runner == "voila":
        yield create_runner_voila(voila_server, notebook_path, page_session, require_vuetify_warmup)
    elif runner == "jupyter_lab":
        yield create_runner_jupyter_lab(jupyter_server, notebook_path, page_session, require_vuetify_warmup)
    elif runner == "jupyter_notebook":
        yield create_runner_jupyter_notebook(jupyter_server, notebook_path, page_session, require_vuetify_warmup)
    else:
        raise RuntimeError(f"Unknown runner {runner}")


@pytest.fixture(scope="session")
def solara_snapshots_directory(request: Any) -> Path:
    path = Path(request.config.rootpath) / "tests" / "ui" / "snapshots"
    if not path.exists():
        path.mkdir(exist_ok=True, parents=True)
    return path


def compare_default(reference, result, threshold=0.1):
    from PIL import Image
    from pixelmatch.contrib.PIL import pixelmatch

    difference = Image.new("RGB", reference.size)
    diff = pixelmatch(reference, result, difference, threshold=threshold)
    return diff, difference


@pytest.fixture
def assert_solara_snapshot(pytestconfig: Any, request: Any, browser_name: str, solara_snapshots_directory) -> Callable:
    from PIL import Image

    testname = f"{str(Path(request.node.name))}".replace("[", "-").replace("]", "").replace(" ", "-").replace(",", "-")
    directory = solara_snapshots_directory / request.node.location[0]
    output_dir = Path(pytestconfig.getoption("--output")) / request.node.location[0]
    if not directory.exists():
        directory.mkdir(exist_ok=True, parents=True)
    if not output_dir.exists():
        output_dir.mkdir(exist_ok=True, parents=True)

    def assert_implementation(
        image: bytes,
        compare: Callable = compare_default,
        testname=testname,
        format="{prefix}{testname}-{platform}{postfix}-{type}.png",
        prefix="",
        postfix="",
    ):
        update_snapshot = pytestconfig.getoption("--solara-update-snapshots")
        format_kwargs = dict(testname=testname, platform=sys.platform, browser=browser_name, prefix=prefix, postfix=postfix)
        path_reference = directory / format.format(**format_kwargs, type="reference").format(**format_kwargs)
        path_reference_output = output_dir / format.format(**format_kwargs, type="reference").format(**format_kwargs)
        path_previous = output_dir / format.format(**format_kwargs, type="previous").format(**format_kwargs)
        path_diff = output_dir / format.format(**format_kwargs, type="diff").format(**format_kwargs)
        if not path_reference.exists():
            if update_snapshot:
                path_reference.write_bytes(image)
            else:
                # CI run, store the reference, but fail
                path_reference_output.write_bytes(image)
                raise AssertionError(
                    f'Snapshot {path_reference} did not exist, file written. Run `cp "{path_reference_output}" "{path_reference}"` '
                    "Commit this file and rerun the CI. Or run with --solara-update-snapshots to update it."
                )
        else:
            if update_snapshot:
                path_reference.write_bytes(image)
            else:
                reference = Image.open(path_reference)
                result = Image.open(BytesIO(image))
                difference = None

                def write():
                    if update_snapshot:
                        path_reference.write_bytes(image)
                    else:
                        # CI run, update the reference in the output dir, and store the previous run next to it
                        path_reference_output.write_bytes(image)
                        reference.save(path_previous)
                        if difference is not None:
                            difference.save(path_diff)

                # the error msg of the default compare is not very helpful
                if reference.size != result.size:
                    write()
                    raise AssertionError(
                        f"Snapshot {path_reference} has a different size than the result {reference.size} != {result.size}."
                        f'Run `cp "{path_reference_output}" "{path_reference}"` Commit this file and rerun the CI. '
                        "Or run with --solara-update-snapshots to update it."
                    )
                diff, difference = compare(reference, result)
                if diff > 0:
                    write()
                    raise AssertionError(
                        f'Snapshot {path_reference} does not match, Run `cp "{path_reference_output}" "{path_reference}"` Commit this file and rerun the CI. '
                        "Or run with --solara-update-snapshots to update it."
                    )

    return assert_implementation


def pytest_addoption(parser: Any) -> None:
    group = parser.getgroup("solara", "Solara")
    group.addoption(
        "--solara-update-snapshots",
        action="store_true",
        default=False,
        help="Do not compare, but store the snapshots.",
    )
    group.addoption(
        "--solara-update-snapshots-ci",
        action="store_true",
        default=False,
        help="On compare failure, store to the reference image. Useful for running in CI and downloading the snapshots.",
    )
    group.addoption(
        "--solara-host",
        default=TEST_HOST,
        help="Host or IP all servers will bind to (solara, jupyter, voila)",
    )
    group.addoption(
        "--solara-port",
        type=int,
        default=TEST_PORT_START + 0,
        help="Port the solara server is running on for the test",
    )
    group.addoption(
        "--jupyter-port",
        type=int,
        default=TEST_PORT_START + 1,
        help="Port the jupyter server is running on for the test (for classic notebook and juptyer lab)",
    )
    group.addoption(
        "--voila-port",
        type=int,
        default=TEST_PORT_START + 2,
        help="Port the voila server is running on for the test (for classic notebook and juptyer lab)",
    )
    vuetify_warmup = os.environ.get("SOLARA_TEST_VUETIFY_WARMUP", "true") in ["true", "True", "1", "on", "On"]
    help = (
        "Load/not load the vuetify fonts and css before running the test, leading to more stable screenshots. If (ipy)vuetify is not used this can be disabled."
    )
    if vuetify_warmup:
        group.addoption(
            "--no-solara-vuetify-warmup",
            action="store_false",
            default=vuetify_warmup,
            help=help,
            dest="solara_vuetify_warmup",
        )
    else:
        group.addoption(
            "--solara-vuetify-warmup",
            action="store_false",
            default=vuetify_warmup,
            help=help,
            dest="solara_vuetify_warmup",
        )



import ipyvuetify as vw

import solara


def get_tick_labels(clazz, min, max, step, **kwargs) -> list:
    el = clazz(label="label", min=min, max=max, step=step, **kwargs)
    _, rc = solara.render(el, handle_error=False)
    switch = rc.find(vw.Slider)
    labels = switch.widget.tick_labels
    rc.close()

    return labels


def test_int_slider():
    reference_ticks = [3, 4, 5]
    assert get_tick_labels(solara.IntSlider, 3, 5, 1, tick_labels="end_points") == ["3", "", "5"]
    assert get_tick_labels(solara.IntSlider, 3, 5, 1, tick_labels=True) == ["3", "4", "5"]
    assert get_tick_labels(solara.IntSlider, 3, 5, 1) is None
    assert get_tick_labels(solara.IntSlider, 3, 5, 1, tick_labels=reference_ticks) == reference_ticks


def test_float_slider():
    reference_ticks = [3.0, 3.5, 4.0, 4.5, 5.0]
    middle_empty_str = [""] * (len(reference_ticks) - 2)
    assert get_tick_labels(solara.FloatSlider, 3, 5, 0.5, tick_labels="end_points") == ["3", *middle_empty_str, "5"]
    assert list(map(float, get_tick_labels(solara.FloatSlider, 3, 5, 0.5, tick_labels=True))) == reference_ticks
    assert get_tick_labels(solara.FloatSlider, 3, 5, 0.5) is None
    assert get_tick_labels(solara.FloatSlider, 3, 5, 0.5, tick_labels=reference_ticks) == reference_ticks

    reference_ticks = [3, 3.3, 3.6, 3.9, 4.2, 4.5, 4.8, 5]
    middle_empty_str = [""] * (len(reference_ticks) - 2)
    assert get_tick_labels(solara.FloatSlider, 3, 5, 0.3, tick_labels="end_points") == ["3", *middle_empty_str, "5"]
    assert get_tick_labels(solara.FloatSlider, 3, 5, 0.3) is None
    assert get_tick_labels(solara.FloatSlider, 3, 5, 0.3, tick_labels=reference_ticks) == reference_ticks



"""
# use_route
"""

import solara
from solara.website.utils import apidoc

title = "use_route"


@solara.component
def Fruit():
    # this gets all routes in fruit's children
    route, routes = solara.use_route()

    if route is None:
        with solara.Link("banana") as main:
            solara.Button("Fruit not found, go to banana")
        return main

    if route.path == "/":
        with solara.Link("banana") as main:
            solara.Button("Choose a fruit, I recomment banana")
        return main

    with solara.Row():
        for route_fruit in routes[1:]:
            with solara.Link(solara.resolve_path(route_fruit)):
                solara.Button(route_fruit.path)

        with solara.Link("/documentation/api/routing/use_route/fruit/nofruit", nofollow=True):
            solara.Button("Wrong fruit")
        with solara.Link("/documentation/api/routing/use_route/not-routed", nofollow=True):
            solara.Button("Wrong url")
    solara.Success(f"You chose {route.path}")


@solara.component
def Page():
    # this gets the top level routes, '/' and 'fruit'
    route_current, routes_all = solara.use_route()
    with solara.Card("Navigation using buttons"):
        with solara.Row():
            for route in routes_all:
                with solara.Link(route):
                    solara.Button(route.path, color="red" if route_current == route else None)
    with solara.Card("Content decided by route:"):
        if route_current is None:
            solara.Error("Page does not exist")
            with solara.Link("fruit/kiwi"):
                solara.Button("Go to fruit/kiwi")
        elif route_current.path == "/":
            with solara.Link("fruit/banana"):
                solara.Button("Go to fruit/banana")
        elif route_current.path == "fruit":
            Fruit()
        else:
            solara.Error(f"Unknown route: {route_current.path}")


routes = [
    solara.Route(path="/"),
    solara.Route(
        path="fruit",
        component=Fruit,
        children=[
            solara.Route(path="/"),
            solara.Route(path="kiwi"),
            solara.Route(path="banana"),
            solara.Route(path="apple"),
        ],
    ),
]

sources = [Fruit.f, Page.f]  # type: ignore
__doc__ += apidoc(solara.use_route)  # type: ignore



import ipyvuetify as v

import solara


def test_sidebar():
    @solara.component
    def Content1():
        def on_click():
            print("click")  # noqa

        with solara.Card("Content") as content:
            with solara.Sidebar():
                solara.Button("Hi1", on_click=on_click)
            solara.Markdown("This is the content1")
        return content

    @solara.component
    def Content2():
        with solara.Card("Content") as content:
            solara.Title("Title2")
            with solara.Sidebar():
                solara.Button("Hi2")
            solara.Markdown("This is the content2")
        return content

    @solara.component
    def Sub3():
        with solara.Card("Sub3") as sub:
            solara.Markdown("Sub3 is the best")
            with solara.Sidebar():
                solara.Button("Hi3b")
        return sub

    @solara.component
    def Content3():
        with solara.Card("Content") as content:
            solara.Title("Title3")
            with solara.Sidebar():
                solara.Button("Hi3")
            solara.Markdown("This is the content3")
            Sub3()
        return content

    @solara.component
    def Content4():
        with solara.Card("Content") as content:
            solara.Title("Title4")
            solara.Markdown("This is the content4")
        return content

    @solara.component
    def Content5():
        with solara.Card("Content") as content:
            with solara.AppBarTitle():
                solara.Markdown("Title4")
                solara.Markdown("Title4b")
                solara.Markdown("Title4c")
            solara.Markdown("This is the content4")
        return content

    set_content = None

    @solara.component
    def Layout():
        nonlocal set_content
        content, set_content = solara.use_state(1)
        with solara.AppLayout(title="Scatter plot") as main:
            if content == 1:
                Content1()
            if content == 2:
                Content2()
            if content == 3:
                Content3()
            if content == 4:
                Content4()
            if content == 5:
                Content5()
        return main

    box, rc = solara.render(Layout(), handle_error=False)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi1"])) == 1
    rc.find(v.ToolbarTitle, children=["Scatter plot"]).assert_not_empty()
    assert set_content is not None
    set_content(2)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi1"])) == 0
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi2"])) == 1
    assert len(rc.find(v.ToolbarTitle, children=["Title2"])) == 1
    set_content(1)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi1"])) == 1
    rc.find(v.ToolbarTitle, children=["Scatter plot"]).assert_not_empty()
    set_content(3)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn)) == 2
    assert len(rc.find(v.ToolbarTitle, children=["Title3"])) == 1
    set_content(4)
    rc.find(v.NavigationDrawer).assert_empty()
    set_content(5)
    assert len(rc.find(v.ToolbarTitle).widget.children) == 3


def test_no_rerender():
    have_sidebar = solara.reactive(False)
    have_app_bar = solara.reactive(False)
    content_use_effects = 0
    content_renders = 0
    app_bar_use_effects = 0
    app_bar_renders = 0

    @solara.component
    def AppBarComponent():
        nonlocal app_bar_renders, app_bar_use_effects

        def effect():
            nonlocal app_bar_use_effects
            app_bar_use_effects += 1

        solara.use_effect(effect, [])
        app_bar_renders += 1
        solara.Button("1")
        solara.Button("2")

    @solara.component
    def Content():
        nonlocal content_renders, content_use_effects

        def effect():
            nonlocal content_use_effects
            content_use_effects += 1

        solara.use_effect(effect, [])
        content_renders += 1
        if have_app_bar.value:
            with solara.AppBar():
                AppBarComponent()

        if have_sidebar.value:
            with solara.Sidebar():
                solara.Button("Hi")
        with solara.Card("Content"):
            solara.Markdown("This is the content")

    @solara.component
    def Layout():
        with solara.AppLayout(title="Test"):
            Content()

    box, rc = solara.render(Layout(), handle_error=False)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi"])) == 0
    assert len(rc.find(v.AppBar).find(v.Btn)) == 0
    assert content_renders == 1
    assert content_use_effects == 1
    assert app_bar_renders == 0
    assert app_bar_use_effects == 0

    have_app_bar.value = True
    assert len(rc.find(v.AppBar).find(v.Btn)) == 2
    assert content_renders == 2
    assert content_use_effects == 1
    assert app_bar_renders == 1
    assert app_bar_use_effects == 1

    have_sidebar.value = True
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi"])) == 1
    assert len(rc.find(v.AppBar).find(v.Btn)) == 2
    assert content_renders == 3
    assert content_use_effects == 1
    assert app_bar_renders == 1
    assert app_bar_use_effects == 1
    rc.close()

    # now, a render with in 1 go
    content_renders = 0
    content_use_effects = 0
    app_bar_renders = 0
    app_bar_use_effects = 0
    box, rc = solara.render(Layout(), handle_error=False)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi"])) == 1
    assert len(rc.find(v.AppBar).find(v.Btn)) == 2
    assert content_renders == 1
    assert content_use_effects == 1
    assert app_bar_renders == 1
    assert app_bar_use_effects == 1



from typing import Optional

import solara

from .head_tag import HeadTag


@solara.component
def Meta(name: Optional[str] = None, property: Optional[str] = None, content: Optional[str] = None):
    """Add a meta tag to the head element, or replace a meta tag with the same name and or property.

    This component should be used inside a [Head](/documentation/components/page/head) component, e.g.:

    ```python
    import solara

    @solara.component
    def Page():
        with solara.VBox() as main:
            MyAwesomeComponent()
            with solara.Head():
                solara.Meta(name="description", property="og:description", content="My page description")
                solara.Meta(property="og:title", content="My page title for social media")
                solara.Meta(property="og:image", content="https://solara.dev/static/assets/images/logo.svg")
                solara.Meta(property="og:type", content="website")
        return main
    ```

    If multiple Meta components are used with the same name+description, the 'deepest' child will take precedence.

    ## Arguments

     * name: The name of the meta tag, used in standard meta tags
     * property: the property of the meta tag, used in Open Graph tags.
     * content: The content of the meta tag.
    """
    attributes = {}
    key = ""
    if name is not None:
        attributes["name"] = name
        key += "-" + name
    if property is not None:
        attributes["property"] = property
        key += "-" + property
    if content is not None:
        attributes["content"] = content
    return HeadTag(tagname="meta", attributes=attributes, key=key)



"""# use_state"""

import solara
import solara.autorouting
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "use_state"
Page = NoPage
__doc__ += apidoc(solara.use_state)  # type: ignore



import pytest

pytest.importorskip("vaex")
from pathlib import Path

import ipyvuetify as v
import ipywidgets as widgets

import solara
import solara.autorouting
import solara.template.portal.solara_portal.pages
import solara.website.pages.documentation
import solara.website.pages.documentation.components.input.button
import solara.website.pages.documentation.getting_started
import solara.widgets
from solara.components.title import TitleWidget
from solara.server.app import AppScript

HERE = Path(__file__).parent


def test_count_arguments():
    def f1(a, b):
        pass

    assert solara.autorouting.count_arguments(f1) == 2

    def f2(a, b, c=1):
        pass

    assert solara.autorouting.count_arguments(f2) == 2

    def f3(a, b, c=None):
        pass

    assert solara.autorouting.count_arguments(f3) == 2

    def f4(a, b, c=None, *args):
        pass

    assert solara.autorouting.count_arguments(f4) == 2

    def f5(a, b, c=None, **kwargs):
        pass

    assert solara.autorouting.count_arguments(f5) == 2


def test_cast():
    def f1(a, b):
        pass

    assert solara.autorouting.arg_cast(["a", "b"], f1) == ["a", "b"]

    def f2(a, b: int):
        pass

    assert solara.autorouting.arg_cast(["a", "42"], f2) == ["a", 42]

    @solara.component
    def F3(a, b: int):
        pass

    assert solara.autorouting.arg_cast(["a", "42"], F3) == ["a", 42]


def test_routes_portal():
    routes = solara.autorouting.generate_routes(solara.template.portal.solara_portal.pages)

    assert routes[0].path == "/"
    assert routes[1].path == "article"
    assert routes[2].path == "tabular"
    assert routes[3].path == "viz"
    assert routes[3].children[0].path == "/"

    main_object = solara.autorouting.RenderPage()
    root = solara.RoutingProvider(routes=routes, pathname="/", children=[main_object])

    container, rc = solara.render(root, handle_error=False)
    nav = rc._find(solara.widgets.Navigator).widget
    title = rc._find(TitleWidget)[-1].widget
    assert title.title == "Solara demo » Home"
    assert rc._find(v.ToolbarTitle).widget.children[0] == "Solara demo » Home"
    nav.location = "/tabular/titanic"
    title = rc._find(TitleWidget)[-1].widget
    assert "titanic" in title.title

    nav.location = "/viz/scatter/titanic"
    title = rc._find(TitleWidget)[-1].widget
    assert "titanic" in title.title
    assert "scatter" in title.title


def test_routes_examples_api_button():
    routes = solara.autorouting.generate_routes(solara.website.pages.documentation.components.input.button)

    assert len(routes) == 1
    assert routes[0].path == "/"

    main_object = solara.autorouting.RenderPage()
    solara_context = solara.RoutingProvider(children=[main_object], routes=routes, pathname="/")

    container, rc = solara.render(solara_context, handle_error=False)

    assert not rc._find(v.NavigationDrawer)


def test_routes_examples_getting_started():
    routes = solara.autorouting.generate_routes(solara.website.pages.documentation.getting_started)

    assert len(routes) == 8
    assert routes[0].path == "/"
    assert routes[0].label == "Quickstart"

    main_object = solara.autorouting.RenderPage()
    solara_context = solara.RoutingProvider(children=[main_object], routes=routes, pathname="/")

    container, rc = solara.render(solara_context, handle_error=False)

    rc._find(v.AppBar).assert_not_empty()


# requires altair as dependency
# def test_routes_examples_examples():

#     routes = solara.autorouting.generate_routes(solara.website.pages)

#     assert len(routes) > 1

#     main_object = solara.autorouting.RenderPage()
#     solara_context = solara.RoutingProvider(children=[main_object], routes=routes, pathname="/")

#     container, rc = solara.render(solara_context, handle_error=False)
#     nav = rc._find(solara.widgets.Navigator).widget
#     nav.location = "/examples/calculator"
#     assert rc._find(v.Tabs, vertical=True)


def test_routes_directory():
    routes = solara.autorouting.generate_routes_directory(HERE / "solara_test_apps" / "multipage")
    assert len(routes) == 8
    assert routes[0].path == "/"
    assert routes[0].label == "Home"

    assert routes[1].path == "my-fruit"
    assert routes[1].label == "My Fruit"
    assert routes[1].children

    assert routes[2].path == "some-markdown"
    assert routes[2].label == "Some Markdown"

    assert routes[3].path == "a-directory"
    assert routes[3].label == "A Directory"
    assert len(routes[3].children) == 2

    assert routes[4].path == "and-notebooks"
    assert routes[4].label == "And Notebooks"

    assert routes[5].path == "custom-routes"
    assert routes[5].label == "Custom Routes"
    assert routes[5].children[0].path == "/"
    assert routes[5].children[0].label == "Hi1"
    assert routes[5].children[1].path == "page2"
    assert routes[5].children[1].label == "Hi2"

    assert routes[6].path == "single-file-directory"
    assert routes[6].label == "Single File Directory"

    assert routes[7].path == "some-other-python-script"
    assert routes[7].label == "Some Other Python Script"

    main_object = solara.autorouting.RenderPage()
    solara_context = solara.RoutingProvider(children=[main_object], routes=routes, pathname="/")

    container, rc = solara.render(solara_context, handle_error=False)
    nav = rc._find(solara.widgets.Navigator).widget
    title = rc._find(TitleWidget).widget
    assert title.title == "Home"

    nav.location = "/my-fruit"
    title = rc._find(TitleWidget).widget
    assert "My Fruit" == title.title

    nav.location = "my-fruit/kiwi"
    title = rc._find(TitleWidget).widget
    assert "My Fruit" == title.title
    template = rc.find(v.Btn).widget.children[0] == "You chose kiwi"

    nav.location = "/some-markdown"
    title = rc._find(TitleWidget).widget
    assert "Some Markdown" == title.title
    template = rc._find(v.VuetifyTemplate)[-1].widget
    assert "renders to highlighted Python code" in template.template

    nav.location = "/some-other-python-script"
    alert = rc._find(v.Alert).widget
    assert "No object with name Page found" in alert.children[0]

    nav.location = "/a-directory"
    title = rc._find(TitleWidget).widget
    assert "Another Markdown" == title.title
    alert = rc._find(v.Alert).widget
    assert "Footer" == alert.children[0]

    nav.location = "/a-directory/not-an-app"
    rc.find(v.Btn, children=["Reset to initial layout"]).assert_single()

    nav.location = "/and-notebooks"
    assert rc._find(v.Slider, label="Language")

    # test navigation in single file directory
    nav.location = "/single-file-directory"
    title = rc._find(TitleWidget).widget
    assert "Single File" == title.title
    assert len(rc._find(v.AppBar)) == 1

    nav.location = "/wrong-path"
    assert "Page not found" in rc._find(v.Alert).widget.children[0]

    nav.location = "/a-directory/wrong-path"
    assert "Page not found" in rc._find(v.Alert).widget.children[0]

    # custom routes in a single file

    nav.location = "/custom-routes"
    button = rc._find(v.Btn, children=["hi1"]).widget
    assert button.children[0] == "hi1"

    nav.location = "/custom-routes/page2"
    button = rc._find(v.Btn, children=["hi2"]).widget
    assert button.children[0] == "hi2"


def test_routes_regular_widgets():
    # routes = solara.autorouting.generate_routes_directory(HERE.parent / "solara_test_apps" / "multipage")
    routes = solara.autorouting.generate_routes_directory(HERE / "solara_test_apps" / "multipage-widgets")

    main_object = solara.autorouting.RenderPage()
    solara_context = solara.RoutingProvider(children=[main_object], routes=routes, pathname="/")

    container, rc = solara.render(solara_context, handle_error=False)
    nav = rc.find(solara.widgets.Navigator).widget

    html = rc.find(v.VuetifyTemplate)[-1].widget
    assert "regular ipywidget" in html.template

    nav.location = "/views"
    rc.find(widgets.Button, description="Never viewed").widget.click()
    assert rc.find(widgets.Button).widget.description == "Viewed 1 times"

    nav.location = "/likes"
    rc.find(widgets.Button, description="No likes recorded").widget.click()
    rc.find(widgets.Button, description="Liked 1 times").widget.click()
    rc.find(widgets.Button, description="Liked 2 times").widget.click()

    # if we navigate back, the state should be preserved for regular ipywidgets
    nav.location = "/views"
    assert rc.find(widgets.Button).widget.description == "Viewed 1 times"

    # but not for elements
    nav.location = "/volume"
    assert rc.find(v.Slider).widget.v_model == 5
    rc.find(v.Slider).widget.v_model = 11
    nav.location = "/views"
    assert rc.find(widgets.Button).widget.description == "Viewed 1 times"
    nav.location = "/volume"
    assert rc.find(v.Slider).widget.v_model == 5


def test_single_file_routes_as_file(kernel_context, no_kernel_context):
    name = str(HERE / "solara_test_apps" / "single_file_routes.py")
    app = AppScript(name)
    app.init()
    assert len(app.routes) == 2
    assert app.routes[0].path == "/"
    assert app.routes[0].layout is not None
    assert app.routes[1].path == "page2"
    assert app.routes[1].layout is None
    try:
        with kernel_context:
            el = app.run()
            root = solara.RoutingProvider(children=[el], routes=app.routes, pathname="/")
            _box, rc = solara.render(root, handle_error=False)
            rc.find(children=["Page 1"]).assert_single()
            rc.find(children=["Custom layout"]).assert_single()
            rc.render(solara.RoutingProvider(children=[el], routes=app.routes, pathname="/page2").key("2"))
            rc.find(children=["Custom layout"]).assert_single()
            rc.find(children=["Page 2"]).assert_single()
    finally:
        app.close()


def test_single_file_routes_as_module(kernel_context, no_kernel_context, extra_include_path):
    with extra_include_path(HERE / "solara_test_apps"):
        app = AppScript("single_file_routes")
        app.init()
        assert len(app.routes) == 2
        assert app.routes[0].path == "/"
        assert app.routes[0].layout is not None
        assert app.routes[1].path == "page2"
        assert app.routes[1].layout is None
        try:
            with kernel_context:
                el = app.run()
                root = solara.RoutingProvider(children=[el], routes=app.routes, pathname="/")
                _box, rc = solara.render(root, handle_error=False)
                rc.find(children=["Page 1"]).assert_single()
                rc.find(children=["Custom layout"]).assert_single()
                rc.render(solara.RoutingProvider(children=[el], routes=app.routes, pathname="/page2").key("2"))
                rc.find(children=["Custom layout"]).assert_single()
                rc.find(children=["Page 2"]).assert_single()
        finally:
            app.close()



import math
import os
from datetime import date, datetime, timedelta
from typing import Callable, List, Optional, Tuple, TypeVar, Union, cast

import ipyvue
import ipyvuetify
import reacton.core
import traitlets
from typing_extensions import Literal

import solara
from solara.alias import rv

T = TypeVar("T")


@solara.value_component(int)
def SliderInt(
    label: str,
    value: Union[int, solara.Reactive[int]] = 0,
    min: int = 0,
    max: int = 10,
    step: int = 1,
    on_value: Optional[Callable[[int], None]] = None,
    thumb_label: Union[bool, Literal["always"]] = True,
    tick_labels: Union[List[str], Literal["end_points"], bool] = False,
    disabled: bool = False,
):
    """Slider for controlling an integer value.

    ### Basic example:

    ```solara
    import solara

    int_value = solara.reactive(42)


    @solara.component
    def Page():
        solara.SliderInt("Some integer", value=int_value, min=-10, max=120)
        solara.Markdown(f"**Int value**: {int_value.value}")
        with solara.Row():
            solara.Button("Reset", on_click=lambda: int_value.set(42))
    ```


    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently selected value.
    * `min`: Minimum value.
    * `max`: Maximum value.
    * `step`: Step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `tick_labels`: Show tick labels corresponding to the values (True),
            custom tick labels by passing a list of strings, only end_points ("end_points"),
            or no labels at all (False, the default).
    * `disabled`: Whether the slider is disabled.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value

    def set_value_cast(value):
        reactive_value.value = int(value)

    updated_tick_labels = _produce_tick_labels(tick_labels, min, max, step)

    return rv.Slider(
        v_model=reactive_value.value,
        on_v_model=set_value_cast,
        label=label,
        min=min,
        max=max,
        step=step,
        thumb_label=thumb_label,
        tick_labels=updated_tick_labels,
        dense=False,
        hide_details=True,
        disabled=disabled,
    )


@solara.value_component(None)
def SliderRangeInt(
    label: str,
    value: Union[Tuple[int, int], solara.Reactive[Tuple[int, int]]] = (1, 3),
    min: int = 0,
    max: int = 10,
    step: int = 1,
    on_value: Callable[[Tuple[int, int]], None] = None,
    thumb_label: Union[bool, Literal["always"]] = True,
    tick_labels: Union[List[str], Literal["end_points"], bool] = False,
    disabled: bool = False,
) -> reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[int, int]]:
    """Slider for controlling a range of integer values.

    ### Basic example:

    ```solara
    import solara

    int_range = solara.reactive((0, 42))


    @solara.component
    def Page():
        solara.SliderRangeInt("Some integer range", value=int_range, min=-10, max=120)
        solara.Markdown(f"**Int range value**: {int_range.value}")
        with solara.Row():
            solara.Button("Reset", on_click=lambda: int_range.set((0, 42)))
    ```

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The currently selected value.
    * `min`: Minimum value.
    * `max`: Maximum value.
    * `step`: Step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `tick_labels`: Show tick labels corresponding to the values (True),
            custom tick labels by passing a list of strings, only end_points ("end_points"),
            or no labels at all (False, the default).
    * `disabled`: Whether the slider is disabled.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value

    def set_value_cast(value):
        v1, v2 = value
        reactive_value.set((int(v1), int(v2)))

    updated_tick_labels = _produce_tick_labels(tick_labels, min, max, step)

    return cast(
        reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[int, int]],
        rv.RangeSlider(
            v_model=reactive_value.value,
            on_v_model=set_value_cast,
            label=label,
            min=min,
            max=max,
            step=step,
            thumb_label=thumb_label,
            tick_labels=updated_tick_labels,
            dense=False,
            hide_details=True,
            disabled=disabled,
        ),
    )


@solara.value_component(float)
def SliderFloat(
    label: str,
    value: Union[float, solara.Reactive[float]] = 0,
    min: float = 0,
    max: float = 10.0,
    step: float = 0.1,
    on_value: Callable[[float], None] = None,
    thumb_label: Union[bool, Literal["always"]] = True,
    tick_labels: Union[List[str], Literal["end_points"], bool] = False,
    disabled: bool = False,
):
    """Slider for controlling a float value.

    ### Basic example:

    ```solara
    import solara

    float_value = solara.reactive(42.4)


    @solara.component
    def Page():
        solara.SliderFloat("Some integer", value=float_value, min=-10, max=120)
        solara.Markdown(f"**Float value**: {float_value.value}")
        with solara.Row():
            solara.Button("Reset", on_click=lambda: float_value.set(42.5))
    ```

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The current value.
    * `min`: The minimum value.
    * `max`: The maximum value.
    * `step`: The step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `tick_labels`: Show tick labels corresponding to the values (True),
            custom tick labels by passing a list of strings, only end_points ("end_points"),
            or no labels at all (False, the default).
    * `disabled`: Whether the slider is disabled.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value

    def set_value_cast(value):
        reactive_value.set(float(value))

    updated_tick_labels = _produce_tick_labels(tick_labels, min, max, step)

    return rv.Slider(
        v_model=reactive_value.value,
        on_v_model=set_value_cast,
        label=label,
        min=min,
        max=max,
        step=step,
        thumb_label=thumb_label,
        tick_labels=updated_tick_labels,
        dense=False,
        hide_details=True,
        disabled=disabled,
    )


@solara.value_component(None)
def SliderRangeFloat(
    label: str,
    value: Union[Tuple[float, float], solara.Reactive[Tuple[float, float]]] = (1.0, 3.0),
    min: float = 0.0,
    max: float = 10.0,
    step: float = 0.1,
    on_value: Callable[[Tuple[float, float]], None] = None,
    thumb_label: Union[bool, Literal["always"]] = True,
    tick_labels: Union[List[str], Literal["end_points"], bool] = False,
    disabled: bool = False,
) -> reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[float, float]]:
    """Slider for controlling a range of float values.

    ### Basic example:

    ```solara
    import solara

    float_range = solara.reactive((0.1, 42.4))


    @solara.component
    def Page():
        solara.SliderRangeFloat("Some float range", value=float_range, min=-10, max=120)
        solara.Markdown(f"**Float range value**: {float_range.value}")
        with solara.Row():
            solara.Button("Reset", on_click=lambda: float_range.set((0.1, 42.4)))
    ```

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The current value.
    * `min`: The minimum value.
    * `max`: The maximum value.
    * `step`: The step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `tick_labels`: Show tick labels corresponding to the values (True),
            custom tick labels by passing a list of strings, only end_points ("end_points"),
            or no labels at all (False, the default).
    * `disabled`: Whether the slider is disabled.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value

    def set_value_cast(value):
        v1, v2 = value
        reactive_value.set((float(v1), float(v2)))

    updated_tick_labels = _produce_tick_labels(tick_labels, min, max, step)

    return cast(
        reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[float, float]],
        rv.RangeSlider(
            v_model=reactive_value.value,
            on_v_model=set_value_cast,
            label=label,
            min=min,
            max=max,
            step=step,
            thumb_label=thumb_label,
            tick_labels=updated_tick_labels,
            dense=False,
            hide_details=True,
            disabled=disabled,
        ),
    )


@solara.value_component(None)
def SliderValue(
    label: str,
    value: Union[T, solara.Reactive[T]],
    values: List[T],
    on_value: Callable[[T], None] = None,
    disabled: bool = False,
) -> reacton.core.ValueElement[ipyvuetify.Slider, T]:
    """Slider for selecting a value from a list of values.

    ### Basic example:

    ```solara
    import solara

    foods = ["Kiwi", "Banana", "Apple"]
    food = solara.reactive("Banana")


    @solara.component
    def Page():
        solara.SliderValue(label="Food", value=food, values=foods)
        solara.Markdown(f"**Selected**: {food.value}")
    ```

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The currently selected value.
    * `values`: List of values to select from.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the slider is disabled.

    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value
    index, set_index = solara.use_state(values.index(reactive_value.value), key="index")

    def on_index(index):
        set_index(index)
        value = values[index]
        reactive_value.set(value)

    return cast(
        reacton.core.ValueElement[ipyvuetify.Slider, T],
        rv.Slider(
            v_model=index,
            on_v_model=on_index,
            ticks=True,
            tick_labels=values,
            label=label,
            min=0,
            max=len(values) - 1,
            dense=False,
            hide_details=True,
            disabled=disabled,
        ),
    )


class DateSliderWidget(ipyvue.VueTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "slider_date.vue"))

    min = traitlets.CFloat(0).tag(sync=True)
    days = traitlets.CFloat(0).tag(sync=True)
    value = traitlets.Any(0).tag(sync=True)

    label = traitlets.Unicode("").tag(sync=True)
    disabled = traitlets.Bool(False).tag(sync=True)


@solara.value_component(date)
def SliderDate(
    label: str,
    value: Union[date, solara.Reactive[date]] = date(2010, 7, 28),
    min: date = date(1981, 1, 1),
    max: date = date(2050, 12, 30),
    on_value: Callable[[date], None] = None,
    disabled: bool = False,
):
    """Slider for controlling a date value.

    ### Basic example:

    ```solara
    import solara
    import datetime

    date_value = solara.reactive(datetime.date(2010, 7, 28))


    @solara.component
    def Page():
        solara.SliderDate("Some date", value=date_value)
        solara.Markdown(f"**Date value**: {date_value.value.strftime('%Y-%b-%d')}")
        with solara.Row():
            solara.Button("Reset", on_click=lambda: date_value.set(datetime.date(2010, 7, 28)))
    ```

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The current value.
    * `min`: The minimum value.
    * `max`: The maximum value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the slider is disabled.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value

    def format(d: date):
        return float(datetime(d.year, d.month, d.day).timestamp())

    dt_min = format(min)
    delta: timedelta = max - min
    days = delta.days

    delta_value: timedelta = reactive_value.value - min
    days_value = delta_value.days
    if days_value < 0:
        days_value = 0

    def set_value_cast(value):
        date = min + timedelta(days=value)
        reactive_value.set(date)

    return DateSliderWidget.element(label=label, min=dt_min, days=days, on_value=set_value_cast, value=days_value, disabled=disabled)


def _produce_tick_labels(tick_labels: Union[List[str], Literal["end_points"], bool], min: float, max: float, step: float) -> Optional[List[str]]:
    if tick_labels == "end_points":
        num_repeats = int(math.ceil((max - min) / step)) - 1
        _tick_labels = [str(min), *([""] * num_repeats), str(max)]
    elif tick_labels is False:
        _tick_labels = None
    elif tick_labels is True:
        _tick_labels, start = [], min

        while start < max:
            _tick_labels.append(str(start))
            start += step
        _tick_labels.append(str(max))
    else:
        _tick_labels = tick_labels

    return _tick_labels


FloatSlider = SliderFloat
IntSlider = SliderInt
ValueSlider = SliderValue
DateSlider = SliderDate



from pathlib import Path

import pandas as pd
import playwright
import playwright.sync_api


def test_download(browser: playwright.sync_api.Browser, page_session: playwright.sync_api.Page, solara_server, solara_app, tmpdir: Path):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/components/output/file_download")
        with page_session.expect_download() as download_info:
            page_session.locator('button:has-text("Download file")').click()
        target = tmpdir / "downloaded.txt"
        download = download_info.value
        download.save_as(target)
        assert target.read_text(encoding="utf8") == "This is the content of the file"

        with page_session.expect_download() as download_info:
            page_session.locator('button:has-text("Download: users.csv")').click()
        target = tmpdir / "downloaded.csv"
        download = download_info.value
        download.save_as(target)
        df = pd.read_csv(target)
        assert df.to_dict() == {"id": {0: 1, 1: 2, 2: 3}, "name": {0: "John", 1: "Mary", 2: "Bob"}}



"""# Input fields

This page contains all the input fields available in Solara.

# InputText
"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

Page = NoPage


__doc__ += apidoc(solara.InputText.f)  # type: ignore
__doc__ += "# InputTextArea"
__doc__ += apidoc(solara.InputTextArea.f)  # type: ignore
__doc__ += "# InputFloat"
__doc__ += apidoc(solara.InputFloat.f)  # type: ignore
__doc__ += "# InputInt"
__doc__ += apidoc(solara.InputInt.f)  # type: ignore

__doc__ += """
# Autofocus Example

```solara
import solara
import solara.lab


@solara.component
def Page():
    show_dialog = solara.use_reactive(False)
    show_conditional = solara.use_reactive(False)
    with solara.Row():
        solara.Button("Show dialog", on_click=lambda: show_dialog.set(True))
        solara.Button("Show conditionally rendered element", on_click=lambda: show_conditional.set(not show_conditional.value))
    with solara.lab.ConfirmationDialog(open=show_dialog):
        solara.InputFloat("Float here", autofocus=True)
    if show_conditional.value:
        solara.InputFloat("Float here", autofocus=True)
```
"""



import solara


@solara.component
def Home():
    with solara.Sidebar():
        solara.SliderInt(label="in sidebar")
    solara.Markdown("Home")


@solara.component
def About():
    solara.Markdown("About")


routes = [
    solara.Route(path="/", component=Home, label="Home"),
    solara.Route(path="about", component=About, label="About"),
]



import ipywidgets as widgets
import playwright.sync_api
import pytest
from IPython.display import display

import solara.util

from .conftest import SERVERS


def test_widget_button_solara(solara_test, page_session: playwright.sync_api.Page, assert_solara_snapshot):
    # this all runs in process, which only works with solara
    # also, this test is only with pure ipywidgets
    button = widgets.Button(description="Click Me!")

    def change_description(obj):
        button.description = "Tested event"

    button.on_click(change_description)
    display(button)
    button_sel = page_session.locator("text=Click Me!")
    button_sel.wait_for()
    button_sel.click()
    page_session.locator("text=Tested event").wait_for()
    # with ipyvuetify 2 we modified the button font for ipywidgets, that does not happen anymore with ipyvuetify 3
    # therefore we have a different screenshot for ipyvuetify 2 and 3
    assert_solara_snapshot(page_session.locator("text=Tested event").screenshot(), postfix="-ipyvuetify3" if solara.util.ipyvuetify_major_version == 3 else "")


def test_widget_ipyaggrid(solara_test, page_session: playwright.sync_api.Page):
    # ipyaggrid >=0.5.2 uses the data-base-url on body
    # introduced in https://github.com/widgetti/ipyaggrid/pull/64
    # which we broke in https://github.com/widgetti/solara/pull/805 since
    # we changed the nbextensions path, but didn't update the data-base-url
    import ipyaggrid

    cars = [
        {"carName": "Chevelle", "origin": "US", "make": "Chevrolet", "price": 30415},
        {"carName": "Skylark 320", "origin": "US", "make": "Buick", "price": 21042},
        {"carName": "PL411", "origin": "Asia", "make": "Datsun", "price": 27676},
    ]
    column_defs = [{"field": c} for c in cars[0]]

    grid_options = {
        "columnDefs": column_defs,
    }
    g = ipyaggrid.Grid(grid_data=cars, grid_options=grid_options)

    watermark_detected = False

    def log(msg):
        if "All AG Grid Enterprise features are unlocked" in msg.text:
            nonlocal watermark_detected
            watermark_detected = True

    page_session.on("console", log)

    display(g)

    cell = page_session.locator(".ag-root-wrapper >> text=Chevrolet")
    cell.click()
    cell.wait_for()
    assert watermark_detected
    page_session.remove_listener("console", log)


def test_solara_button_all(ipywidgets_runner, page_session: playwright.sync_api.Page, request, assert_solara_snapshot):
    if request.node.callspec.params["ipywidgets_runner"] != "solara" and request.node.callspec.params["solara_server"] != SERVERS[0]:
        pytest.skip("No need to run this test for all servers.")

    # this function (or rather its lines) will be executed in the kernel
    # voila, lab, classic notebook and solara will all execute it
    def kernel_code():
        import solara

        @solara.component
        def Button():
            text, set_text = solara.use_state("Click Me!")

            def on_click():
                set_text("Tested event")

            with solara.Row():
                solara.Button(text, on_click=on_click)

        display(Button())

    ipywidgets_runner(kernel_code)
    button_sel = page_session.locator("button >> text=Click Me!")
    assert_solara_snapshot(button_sel.screenshot())
    button_sel.wait_for()
    button_sel.click()
    page_session.locator("button >> text=Tested event").wait_for()
    page_session.wait_for_timeout(1000)


def test_slider_all(ipywidgets_runner, page_session: playwright.sync_api.Page, request, assert_solara_snapshot):
    if request.node.callspec.params["ipywidgets_runner"] != "solara" and request.node.callspec.params["solara_server"] != SERVERS[0]:
        pytest.skip("No need to run this test for all servers.")

    # this function (or rather its lines) will be executed in the kernel
    # voila, lab, classic notebook and solara will all execute it
    def kernel_code():
        import ipywidgets as widgets

        slider = widgets.FloatSlider(value=7.5, min=5.0, max=10.0, step=0.1, description="FloatSlider")
        box = widgets.HBox([slider])
        box.add_class("my-slider")
        # we do not want it full with, which depends on environment/browser window size
        box.layout.width = "400px"
        display(box)

    ipywidgets_runner(kernel_code)
    slider_sel = page_session.locator(".my-slider")
    assert_solara_snapshot(slider_sel.screenshot(), postfix="-ipywidgets-" + str(solara.util.ipywidgets_major))
    page_session.wait_for_timeout(1000)



"""# Meta"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    solara.Info("Nothing to see here, only in this page's source code, or by looking at the google search results for this page.")
    with solara.Head():
        solara.Meta(
            name="description",
            content="The Meta component can be used to set the description of a page. This is useful for SEO, or crawlers that index your page.",
        )


__doc__ += apidoc(solara.Meta.f)  # type: ignore



import sys
from typing import List
import webbrowser
from qtpy.QtWidgets import QApplication
from qtpy.QtWebEngineWidgets import QWebEngineView
from qtpy.QtWebChannel import QWebChannel
from qtpy import QtCore, QtGui
import signal
from pathlib import Path

HERE = Path(__file__).parent


# setUrlRequestInterceptor, navigationRequested and acceptNavigationRequest
# all trigger the websocket to disconnect, so we need to block cross origin
# requests on the frontend/browser side by intercepting clicks on links

cross_origin_block_js = """
var script = document.createElement('script');
script.src = 'qrc:///qtwebchannel/qwebchannel.js';
document.head.appendChild(script);
script.onload = function() {
    new QWebChannel(qt.webChannelTransport, function(channel) {
        let py_callback = channel.objects.py_callback;

        document.addEventListener('click', function(event) {
            let target = event.target;
            while (target && target.tagName !== 'A') {
                target = target.parentNode;
            }

            if (target && target.tagName === 'A') {
                const linkOrigin = new URL(target.href).origin;
                const currentOrigin = window.location.origin;

                if (linkOrigin !== currentOrigin) {
                    event.preventDefault();
                    console.log("Blocked cross-origin navigation to:", target.href);
                    py_callback.open_link(target.href);  // Call Python method
                }
            }
        }, true);
    });
};
"""


class PyCallback(QtCore.QObject):
    @QtCore.Slot(str)
    def open_link(self, url):
        webbrowser.open(url)


class QWebEngineViewWithPopup(QWebEngineView):
    # keep a strong reference to all windows
    windows: List = []

    def __init__(self):
        super().__init__()
        self.page().newWindowRequested.connect(self.handle_new_window_request)

        # Set up WebChannel and py_callback object
        self.py_callback = PyCallback()
        self.channel = QWebChannel()
        self.channel.registerObject("py_callback", self.py_callback)
        self.page().setWebChannel(self.channel)

        self.loadFinished.connect(self._inject_javascript)

    def _inject_javascript(self, ok):
        self.page().runJavaScript(cross_origin_block_js)

    def handle_new_window_request(self, info):
        webview = QWebEngineViewWithPopup()
        geometry = info.requestedGeometry()
        webview.resize(geometry.width(), geometry.height())
        webview.setUrl(info.requestedUrl())
        webview.show()
        QWebEngineViewWithPopup.windows.append(webview)
        return webview


def run_qt(url):
    app = QApplication(["Solara App"])
    web = QWebEngineViewWithPopup()
    web.setUrl(QtCore.QUrl(url))
    web.resize(1024, 1024)
    web.show()

    app_name = "Solara"
    app.setApplicationDisplayName(app_name)
    app.setApplicationName(app_name)
    web.setWindowTitle(app_name)
    app.setWindowIcon(QtGui.QIcon(str(HERE.parent / "website/public/logo.svg")))
    if sys.platform.startswith("darwin"):
        # Set app name, if PyObjC is installed
        # Python 2 has PyObjC preinstalled
        # Python 3: pip3 install pyobjc-framework-Cocoa
        try:
            from Foundation import NSBundle

            bundle = NSBundle.mainBundle()
            if bundle:
                app_info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
                if app_info is not None:
                    app_info["CFBundleName"] = app_name
                    app_info["CFBundleDisplayName"] = app_name
        except ModuleNotFoundError:
            pass

    # without this, ctrl-c does not work in the terminal
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app.exec_()



"""# Authorization

Authorization is a common requirement for web applications. This example shows how to implement a simple login form and
how to use `use_route` to implement authorization.

The `Layout` component checks if the current route requires authorization and if the user is logged in. If not, it
redirects to the login form.
"""

import dataclasses
from typing import Optional, cast

import solara
import solara.lab

github_url = solara.util.github_url(__file__)

route_order = ["/", "users", "admin"]


def check_auth(route, children):
    # This can be replaced by a custom function that checks if the user is
    # logged in and has the required permissions.

    # routes that are public or only for admin
    # the rest only requires login
    public_paths = ["/"]
    admin_paths = ["admin"]

    if route.path in public_paths:
        children_auth = children
    else:
        if user.value is None:
            children_auth = [LoginForm()]
        else:
            if route.path in admin_paths and not user.value.admin:
                children_auth = [solara.Error("You are not an admin")]
            else:
                children_auth = children
    return children_auth


@dataclasses.dataclass
class User:
    username: str
    admin: bool = False


user = solara.reactive(cast(Optional[User], None))
login_failed = solara.reactive(False)


def login(username: str, password: str):
    # this function can be replace by a custom username/password check
    if username == "test" and password == "test":
        user.value = User(username, admin=False)
        login_failed.value = False
    elif username == "admin" and password == "admin":
        user.value = User(username, admin=True)
        login_failed.value = False
    else:
        login_failed.value = True


@solara.component
def Page():
    solara.Markdown("This page is visible for everyone")

    solara.Markdown(__doc__)
    solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)


@solara.component
def LoginForm():
    username = solara.use_reactive("")
    password = solara.use_reactive("")
    with solara.Card("Login"):
        solara.Markdown(
            """
        This is an example login form.

          * use admin/admin to login as admin.
          * use test/test to login as a normal user.
        """
        )
        solara.InputText(label="Username", value=username)
        solara.InputText(label="Password", password=True, value=password)
        solara.Button(label="Login", on_click=lambda: login(username.value, password.value))
        if login_failed.value:
            solara.Error("Wrong username or password")


@solara.component
def Layout(children=[]):
    route, routes = solara.use_route(peek=True)
    if route is None:
        return solara.Error("Route not found")

    children = check_auth(route, children)

    with solara.AppLayout(children=children, navigation=True):
        with solara.AppBar():
            with solara.lab.Tabs(align="center"):
                for route in routes:
                    name = route.path if route.path != "/" else "Home"
                    is_admin = user.value and user.value.admin
                    # we could skip the admin tab if the user is not an admin
                    # if route.path == "admin" and not is_admin:
                    #     continue
                    # in this case we disable the tab
                    disabled = route.path == "admin" and not is_admin
                    solara.lab.Tab(name, path_or_route=route, disabled=disabled)
            if user.value:
                solara.Text(f"Logged in as {user.value.username} as {'admin' if user.value.admin else 'user'}")
                with solara.Tooltip("Logout"):
                    solara.Button(icon_name="mdi-logout", icon=True, on_click=lambda: user.set(None))
            else:
                with solara.AppBar():
                    solara.Text("Not logged in")



import solara


@solara.component
def Page():
    x, set_x = solara.use_state(2)
    x_squared = x**2

    with solara.Sidebar():
        solara.Markdown("## My First Solara app ☀️")
        solara.SliderInt(label="x", value=x, on_value=set_x)
    solara.Markdown(f"{x} squared = {x_squared}")


@solara.component
def Layout(children):
    route, routes = solara.use_route()
    return solara.AppLayout(children=children)



"""# CrossFilterReport"""

import plotly

import solara
import solara.lab
from solara.website.utils import apidoc

df = plotly.data.gapminder()


@solara.component
def Page():
    solara.provide_cross_filter()
    solara.CrossFilterReport(df, classes=["py-2"])
    solara.CrossFilterSelect(df, "country")
    solara.CrossFilterSlider(df, "gdpPercap", mode=">")


__doc__ += apidoc(solara.CrossFilterReport.f)  # type: ignore



from typing import List

import reacton
import solara


@reacton.component
def Head(children: List[reacton.core.Element] = []):
    """A component that manager the "head" tag of the page to avoid duplicate tags, such as titles.

    Currently only supports the [title](/documentation/components/page/title) tag as child, e.g.:

    ```python
    import solara

    @solara.component
    def Page():
        with solara.VBox() as main:
            MyAwesomeComponent()
            with solara.Head():
                solara.Title("My page title")
        return main

    ```

    """
    return solara.Div(children=children, style="display; none")



from typing import List, Union


def get_pandas_major():
    import pandas as pd

    return int(pd.__version__[0])


def df_type(df):
    return df.__class__.__module__.split(".")[0]


def df_len(df) -> int:
    """Return the number of rows in a dataframe."""
    return len(df)


def df_columns(df) -> List[str]:
    """Return a list of column names from a dataframe."""
    if df_type(df) == "vaex":
        return df.get_column_names()
    elif df_type(df) == "pandas":
        return df.columns.tolist()
    elif df_type(df) == "polars":
        return df.columns
    else:
        raise TypeError(f"{type(df)} not supported")


def df_row_names(df) -> List[Union[int, str]]:
    """Return a list of row names from a dataframe."""
    if df_type(df) == "vaex" or df_type(df) == "polars":
        return list(range(df_len(df)))
    elif df_type(df) == "pandas":
        return df.index.tolist()
    else:
        raise TypeError(f"{type(df)} not supported")


def df_slice(df, start: int, stop: int):
    """Return a subset of rows from a dataframe."""
    if df_type(df) == "pandas":
        return df.iloc[start:stop]
    else:
        return df[start:stop]


def df_records(df) -> List[dict]:
    """A list of records from a dataframe."""
    if df_type(df) == "pandas":
        return df.to_dict("records")
    elif df_type(df) == "polars":
        return df.to_dicts()
    elif df_type(df) == "vaex":
        return df.to_records()
    else:
        raise TypeError(f"{type(df)} not supported")


def df_unique(df, column, limit=None):
    if df_type(df) == "vaex":
        return df.unique(column, limit=limit + 1 if limit else None, limit_raise=False)
    if df_type(df) == "pandas":
        x = df[column].unique()  # .to_numpy()
        return x[:limit]
    else:
        raise TypeError(f"{type(df)} not supported")


def df_value_count(df, column, limit=None):
    if df_type(df) == "vaex":
        dfv = df.groupby(column, agg="count", sort="count", ascending=False)
        dfv = dfv.to_pandas_df().rename({column: "value"}, axis=1)
        return dfv[:limit]
    if df_type(df) == "pandas":
        dfv = df[column].value_counts(dropna=False).to_frame()
        dfv = dfv.reset_index()
        if get_pandas_major() >= 2:
            dfv = dfv.rename({column: "value"}, axis=1)
        else:
            dfv = dfv.rename({"index": "value", column: "count"}, axis=1)
        return dfv[:limit]
    else:
        raise TypeError(f"{type(df)} not supported")


def df_range(df, column):
    if df_type(df) == "vaex":
        minmax = df[column].minmax()
        return minmax[0].item(), minmax[1].item()
    if df_type(df) == "pandas":
        return df[column].min().item(), df[column].max().item()
    else:
        raise TypeError(f"{type(df)} not supported")


def df_filter_missing(df, column):
    if df_type(df) == "vaex":
        return df[column].isna()
    if df_type(df) == "pandas":
        return df[column].isna()
    else:
        raise TypeError(f"{type(df)} not supported")


def df_filter_values(df, column, values, invert=False):
    if df_type(df) == "vaex":
        filter = df[column].isin(values)
        if invert:
            filter = ~filter
        return filter
    if df_type(df) == "pandas":
        filter = df[column].isin(values)
        if invert:
            filter = ~filter
        return filter
    else:
        raise TypeError(f"{type(df)} not supported")


def df_py_types(df):
    """Return a dict with column names as keys and python types as values.

    Support types are
        * int
        * float
        * str
        * bool

    If a type is not supported, the internal type is returned.

    """
    import numpy as np

    if df_type(df) == "vaex":
        schema = df.schema()
        py_types = [int, float, str, bool]

        def py_type(dtype):
            for k in py_types:
                if dtype == k:
                    return k
            return dtype

        return {name: py_type(dtype) for name, dtype in schema.items()}
    elif df_type(df) == "pandas":
        schema = df.dtypes

        def py_type(dtype):
            if isinstance(dtype, np.dtype):
                if dtype.kind in "iu":
                    return int
                elif dtype.kind == "f":
                    return float
                elif dtype.kind == "b":
                    return bool
                else:
                    return dtype
            else:
                return dtype

        return {name: py_type(dtype) for name, dtype in schema.items()}
    else:
        raise TypeError(f"{type(df)} not supported")



import solara
from solara import autorouting
from solara.alias import rv
from solara.components.title import Title
from solara.server import server
from solara.website.components.algolia import Algolia

from ..components import Header

title = "Home"

route_order = ["/", "showcase", "documentation", "apps", "contact", "changelog", "roadmap", "pricing", "our_team", "careers", "about", "scale_ipywidgets"]


_redirects = {
    "/docs": "/documentation/getting_started/introduction",
    "/docs/installing": "/documentation/getting_started/installing",
    "/docs/quickstart": "/documentation/getting_started",
    "/docs/tutorial": "/documentation/getting_started/tutorials",
    "/docs/tutorial/data-science": "/documentation/getting_started/tutorials/data-science",
    "/docs/tutorial/web-app": "/documentation/getting_started/tutorials/web-app",
    "/docs/tutorial/ipywidgets": "/documentation/getting_started/tutorials/ipywidgets",
    "/docs/tutorial/streamlit": "/documentation/getting_started/tutorials/streamlit",
    "/docs/tutorial/dash": "/documentation/getting_started/tutorials/dash",
    "/docs/tutorial/jupyter-dashboard-part1": "/documentation/getting_started/tutorials/jupyter-dashboard-part1",
    "/docs/fundamentals": "/documentation/getting_started/fundamentals",
    "/docs/fundamentals/components": "/documentation/getting_started/fundamentals/components",
    "/docs/fundamentals/state-management": "/documentation/getting_started/fundamentals/state-management",
    "/docs/howto": "/documentation/advanced/howto",
    "/docs/howto/contribute": "/documentation/advanced/development/contribute",
    "/docs/howto/multipage": "/documentation/advanced/howto/multipage",
    "/docs/howto/layout": "/documentation/advanced/howto/layout",
    "/docs/howto/testing": "/documentation/advanced/howto/testing",
    "/docs/howto/debugging": "/documentation/advanced/howto/debugging",
    "/docs/howto/embed": "/documentation/advanced/howto/embed",
    "/docs/howto/ipywidget-libraries": "/documentation/advanced/howto/ipywidget-libraries",
    "/docs/reference": "/documentation/advanced/reference",
    "/docs/reference/static-files": "/documentation/advanced/reference/static-files",
    "/docs/reference/asset-files": "/documentation/advanced/reference/asset-files",
    "/docs/reference/static-site-generation": "/documentation/advanced/reference/static-site-generation",
    "/docs/reference/search": "/documentation/advanced/reference/search",
    "/docs/reference/reloading": "/documentation/advanced/reference/reloading",
    "/docs/reference/notebook-support": "/documentation/advanced/reference/notebook-support",
    "/docs/reference/caching": "/documentation/advanced/reference/caching",
    "/docs/understanding": "/documentation/advanced/understanding",
    "/docs/understanding/ipywidgets": "/documentation/advanced/understanding/ipywidgets",
    "/docs/understanding/ipyvuetify": "/documentation/advanced/understanding/ipyvuetify",
    "/docs/understanding/reacton": "/documentation/advanced/understanding/reacton",
    "/docs/understanding/reacton-basics": "/documentation/advanced/understanding/reacton-basics",
    "/docs/understanding/anatomy": "/documentation/advanced/understanding/anatomy",
    "/docs/understanding/rules-of-hooks": "/documentation/advanced/understanding/rules-of-hooks",
    "/docs/understanding/containers": "/documentation/advanced/understanding/containers",
    "/docs/understanding/solara": "/documentation/advanced/understanding/solara",
    "/docs/understanding/routing": "/documentation/advanced/understanding/routing",
    "/docs/understanding/solara-server": "/documentation/advanced/understanding/solara-server",
    "/docs/understanding/voila": "/documentation/advanced/understanding/voila",
    "/docs/deploying": "/documentation/getting_started/deploying",
    "/docs/deploying/self-hosted": "/documentation/getting_started/deploying/self-hosted",
    "/docs/deploying/cloud-hosted": "/documentation/getting_started/deploying/cloud-hosted",
    "/docs/enterprise": "/documentation/advanced/enterprise",
    "/docs/enterprise/oauth": "/documentation/advanced/enterprise/oauth",
    "/docs/development": "/documentation/advanced/development/setup",
    "/docs/troubleshoot": "/documentation/getting_started/troubleshoot",
    "/docs/changelog": "/changelog",
    "/docs/contact": "/contact",
    "/docs/faq": "/documentation/faq",
    "/docs/lab": "/documentation/getting_started/what-is-lab",
    "/api": "/documentation/api",
    "/api/altair": "/documentation/components/viz/altair",
    "/api/app_bar": "/documentation/components/layout/app_bar",
    "/api/app_bar_title": "/documentation/components/layout/app_bar_title",
    "/api/app_layout": "/documentation/components/layout/app_layout",
    "/api/avatar": "/documentation/components/enterprise/avatar",
    "/api/avatar_menu": "/documentation/components/enterprise/avatar_menu",
    "/api/card": "/documentation/components/layout/card",
    "/api/card_actions": "/documentation/components/layout/card_actions",
    "/api/column": "/documentation/components/layout/column",
    "/api/columns": "/documentation/components/layout/columns",
    "/api/columns_responsive": "/documentation/components/layout/columns_responsive",
    "/api/cross_filter_dataframe": "/documentation/api/cross_filter/cross_filter_dataframe",
    "/api/cross_filter_report": "/documentation/api/cross_filter/cross_filter_report",
    "/api/cross_filter_select": "/documentation/api/cross_filter/cross_filter_select",
    "/api/cross_filter_slider": "/documentation/api/cross_filter/cross_filter_slider",
    "/api/generate_routes": "/documentation/api/routing/generate_routes",
    "/api/generate_routes_directory": "/documentation/api/routing/generate_routes_directory",
    "/api/resolve_path": "/documentation/api/routing/resolve_path",
    "/api/resolve_path/kiwi": "/documentation/api/routing/resolve_path/kiwi",
    "/api/resolve_path/banana": "/documentation/api/routing/resolve_path/banana",
    "/api/resolve_path/apple": "/documentation/api/routing/resolve_path/apple",
    "/api/link": "/documentation/components/advanced/link",
    "/api/link/kiwi": "/documentation/components/advanced/link/kiwi",
    "/api/link/banana": "/documentation/components/advanced/link/banana",
    "/api/link/apple": "/documentation/components/advanced/link/apple",
    "/api/use_route": "/documentation/api/routing/use_route",
    "/api/use_route/fruit": "/documentation/api/routing/use_route/fruit",
    "/api/use_route/fruit/kiwi": "/documentation/api/routing/use_route/fruit/kiwi",
    "/api/use_route/fruit/banana": "/documentation/api/routing/use_route/fruit/banana",
    "/api/use_route/fruit/apple": "/documentation/api/routing/use_route/fruit/apple",
    "/api/use_router": "/documentation/api/routing/use_router",
    "/api/use_cross_filter": "/documentation/api/hooks/use_cross_filter",
    "/api/use_dark_effective": "/documentation/api/hooks/use_dark_effective",
    "/api/use_effect": "/documentation/api/hooks/use_effect",
    "/api/use_exception": "/documentation/api/hooks/use_exception",
    "/api/use_memo": "/documentation/api/hooks/use_memo",
    "/api/use_previous": "/documentation/api/hooks/use_previous",
    "/api/use_reactive": "/documentation/api/hooks/use_reactive",
    "/api/use_state": "/documentation/api/hooks/use_state",
    "/api/use_state_or_update": "/documentation/api/hooks/use_state_or_update",
    "/api/use_task": "/documentation/components/lab/use_task",
    "/api/use_thread": "/documentation/api/hooks/use_thread",
    "/api/use_trait_observe": "/documentation/api/hooks/use_trait_observe",
    "/examples/fullscreen": "/documentation/examples/fullscreen",
    "/examples/fullscreen/authorization": "/apps/authorization",
    "documentation/examples/fullscreen/authorization": "/apps/authorization",
    "/examples/fullscreen/layout-demo": "/apps/layout-demo",
    "/documentation/examples/fullscreen/layout_demo": "/apps/layout-demo",
    "/examples/fullscreen/multipage": "/apps/multipage",
    "/documentation/examples/fullscreen/multipage": "/apps/multipage",
    "/examples/fullscreen/scatter": "apps/scatter",
    "/documentation/examples/fullscreen/scatter": "/apps/scatter",
    "/examples/fullscreen/scrolling": "/apps/scrolling",
    "/documentation/examples/fullscreen/scrolling": "/apps/scrolling",
    "/examples/fullscreen/tutorial-streamlit": "/apps/tutorial-streamlit",
    "/documentation/examples/fullscreen/tutorial_streamlit": "/apps/tutorial-streamlit",
    "/api/route": "/documentation/api/routing/route",
    "/api/route/kiwi": "/documentation/api/routing/route/kiwi",
    "/api/route/banana": "/documentation/api/routing/route/banana",
    "/api/route/apple": "/documentation/api/routing/route/apple",
    "/examples/libraries": "/documentation/examples/libraries",
    "/examples/libraries/altair": "/documentation/examples/libraries/altair",
    "/examples/libraries/bqplot": "/documentation/examples/libraries/bqplot",
    "/examples/libraries/ipyleaflet": "/documentation/examples/libraries/ipyleaflet",
    "/examples/libraries/ipyleaflet_advanced": "/documentation/examples/libraries/ipyleaflet_advanced",
    "/examples/utilities": "/documentation/examples/utilities",
    "/examples/utilities/calculator": "/documentation/examples/utilities/calculator",
    "/examples/utilities/countdown_timer": "/documentation/examples/utilities/countdown_timer",
    "/examples/utilities/todo": "/documentation/examples/utilities/todo",
    "/examples/visualization": "/documentation/examples/visualization",
    "/examples/visualization/annotator": "/documentation/examples/visualization/annotator",
    "/examples/visualization/linked_views": "/documentation/examples/visualization/linked_views",
    "/examples/visualization/plotly": "/documentation/examples/visualization/plotly",
    "/examples/general": "/documentation/examples/general",
    "/examples/general/custom_storage": "/documentation/examples/general/custom_storage",
    "/examples/general/deploy_model": "/documentation/examples/general/deploy_model",
    "/examples/general/live_update": "/documentation/examples/general/live_update",
    "/examples/general/login_oauth": "/documentation/examples/general/login_oauth",
    "/examples/general/pokemon_search": "/documentation/examples/general/pokemon_search",
    "/examples/general/vue_component": "/documentation/examples/general/vue_component",
    "/examples": "/documentation/examples",
    "/examples/ai": "/documentation/examples/ai",
    "/examples/ai/chatbot": "/documentation/examples/ai/chatbot",
    "/examples/ai/tokenizer": "/documentation/examples/ai/tokenizer",
    "/examples/basics": "/documentation/examples/basics",
    "/examples/basics/sine": "/documentation/examples/basics/sine",
    "/api/get_kernel_id": "/documentation/api/utilities/get_kernel_id",
    "/api/get_session_id": "/documentation/api/utilities/get_session_id",
    "/api/on_kernel_start": "/documentation/api/utilities/on_kernel_start",
    "/api/component_vue": "/documentation/api/utilities/component_vue",
    "/api/plotly": "/documentation/components/viz/plotly",
    "/api/plotly_express": "/documentation/components/viz/plotly_express",
    "/api/tab": "/documentation/components/lab/tab",
    "/api/tabs": "/documentation/components/lab/tabs",
    "/api/task": "/documentation/components/lab/task",
    "/api/computed": "/documentation/api/utilities/computed",
    "/api/markdown": "/documentation/components/output/markdown",
    "/api/markdown_editor": "/documentation/components/output/markdown_editor",
    "/api/matplotlib": "/documentation/components/viz/matplotlib",
    "/api/echarts": "/documentation/components/viz/echarts",
    "/api/theming": "/documentation/components/lab/theming",
    "/api/griddraggable": "/documentation/components/layout/griddraggable",
    "/api/gridfixed": "/documentation/components/layout/gridfixed",
    "/api/html": "/documentation/components/output/html",
    "/api/file_download": "/documentation/components/output/file_download",
    "/api/hbox": "/documentation/components/layout/hbox",
    "/api/vbox": "/documentation/components/layout/vbox",
    "/api/sidebar": "/documentation/components/layout/sidebar",
    "/api/row": "/documentation/components/layout/row",
    "/api/error": "/documentation/components/status/error",
    "/api/info": "/documentation/components/status/info",
    "/api/progress": "/documentation/components/status/progress",
    "/api/spinner": "/documentation/components/status/spinner",
    "/api/success": "/documentation/components/status/success",
    "/api/warning": "/documentation/components/status/warning",
    "/api/image": "/documentation/components/output/image",
    "/api/sql_code": "/documentation/components/output/sql_code",
    "/api/tooltip": "/documentation/components/output/tooltip",
    "/api/chat": "/documentation/components/lab/chat",
    "/api/confirmation_dialog": "/documentation/components/lab/confirmation_dialog",
    "/api/cookies_headers": "/documentation/components/lab/cookies_headers",
    "/api/input_date": "/documentation/components/lab/input_date",
    "/api/button": "/documentation/components/input/button",
    "/api/checkbox": "/documentation/components/input/checkbox",
    "/api/file_browser": "/documentation/components/input/file_browser",
    "/api/file_drop": "/documentation/components/input/file_drop",
    "/api/meta": "/documentation/components/advanced/meta",
    "/api/style": "/documentation/components/advanced/style",
    "/api/menu": "/documentation/components/lab/menu",
    "/api/head": "/documentation/components/page/head",
    "/api/input": "/documentation/components/input/input",
    "/api/select": "/documentation/components/input/select",
    "/api/slider": "/documentation/components/input/slider",
    "/api/switch": "/documentation/components/input/switch",
    "/api/togglebuttons": "/documentation/components/input/togglebuttons",
    "/api/dataframe": "/documentation/components/data/dataframe",
    "/api/pivot_table": "/documentation/components/data/pivot_table",
    "/api/display": "/documentation/api/utilities/display",
    "/api/memoize": "/documentation/api/utilities/memoize",
    "/api/reactive": "/documentation/api/utilities/reactive",
    "/api/widget": "/documentation/api/utilities/widget",
    "/api/default_layout": "/documentation/components/layout",
    "/api/title": "/documentation/components/page/title",
}


server._redirects = _redirects
autorouting._redirects = _redirects


@solara.component
def Page():
    solara.Markdown("should not see me")


@solara.component
def SimpleListItem(text=None, children=[], class_: str = None, icon_name=None):
    if icon_name is not None:
        children = [rv.Icon(children=[icon_name])] + children
    if text is not None:
        children = [text] + children
    return rv.Html(tag="li", children=children, attributes={"class": class_})


@solara.component
def List(children=[], class_: str = None):
    return rv.Html(tag="ul", children=children, attributes={"class": class_})


@solara.component_vue("home.vue")
def Home(children=[]):
    pass


@solara.component
def Layout(children=[]):
    router = solara.use_router()
    route_current, all_routes = solara.use_route()
    route_sidebar_current, all_routes_sidebar = solara.use_route(1)

    show_left_menu, set_show_left_menu = solara.use_state(False)
    show_right_menu, set_show_right_menu = solara.use_state(False)

    target, set_target = solara.use_state(0)

    if route_current is not None:
        with solara.Head():
            solara.Meta(name="twitter:card", content="summary_large_image")
            solara.Meta(name="twitter:site", content="@solara_dev")
            solara.Meta(name="twitter:image", content="https://solara.dev/static/assets/images/logo-small.png")
            solara.Meta(property="og:url", content="https://solara.dev" + router.path)
            solara.Meta(property="og:image", content="https://solara.dev/static/assets/images/logo-small.png")
            solara.Meta(property="og:type", content="website")
    if route_current is not None and route_current.path == "apps":
        return children[0]
    elif route_current is not None and route_current.path == "/":
        Title(title="Solara: Build high-quality web applications in pure Python")
        Home()
    else:
        with solara.VBox(grow=False):
            Title(title="Solara documentation")
            Header(
                on_toggle_left_menu=lambda: set_show_left_menu(not show_left_menu),
                on_toggle_right_menu=lambda: set_show_right_menu(not show_right_menu),
            )
            with rv.Container(tag="section", fluid=True, ma_0=True, pa_0=True, class_="fill-height solara-content-main"):
                if route_current is None:
                    return solara.Error("Page not found")
                else:
                    with rv.Row(
                        style_="flex-wrap: nowrap; margin: 0; min-height: calc(100vh - 64px);",
                        justify="center" if route_current is not None and route_current.path in ["documentation", "showcase"] else "start",
                    ):
                        if route_current is not None and route_current.module is not None and hasattr(route_current.module, "Sidebar"):
                            with solara.v.Sheet(
                                style_="""
                                    height: 100vh;
                                    width: 20rem;
                                    overflow: auto;
                                    border-right: 1px solid var(--color-border-appbar);
                                    position: sticky;
                                    top: 0;
                                    flex-direction: column;
                                    gap: 0;
                                """,
                                class_="d-md-flex d-none",
                                elevation=0,
                            ):
                                route_current.module.Sidebar()
                        with rv.Col(
                            tag="main",
                            md=True,
                            class_="pa-0",
                            style_="max-width: 1024px" if route_current.path == "showcase" else "",
                        ):
                            with solara.Row(
                                children=children,
                                justify="center",
                                classes=["solara-page-content-search"],
                                style="height: unset",
                            ):
                                pass

                # absolute = True prevents the drawer from being below the overlay it generates
                # Drawer navigation for top menu
                with rv.NavigationDrawer(
                    v_model=show_right_menu,
                    on_v_model=set_show_right_menu,
                    fixed=True,
                    absolute=True,
                    right=True,
                    hide_overlay=False,
                    overlay_color="#000000",
                    overlay_opacity=0.5,
                    style_="height: 100vh",
                ):
                    Algolia()
                    with rv.List(nav=True):
                        with rv.ListItemGroup(active_class="text--primary"):
                            for route in all_routes:
                                if route.path == "apps":
                                    continue
                                with solara.Link(route):
                                    solara.ListItem(route.label)

                if route_current is not None and route_current.module is not None and hasattr(route_current.module, "Sidebar"):
                    with solara.v.NavigationDrawer(
                        absolute=True,
                        clipped=True,
                        class_="d-md-none d-block",
                        height="unset",
                        style_="min-height: 100vh;",
                        v_model=show_left_menu,
                        on_v_model=set_show_left_menu,
                        width="20rem",
                    ):
                        route_current.module.Sidebar()


@solara.component
def Testimonial(text, name, position, img):
    max_width = "300px"
    with rv.Card(
        elevation=2,
        dark=False,
        max_width=max_width,
        class_="testimonial-card",
    ):
        with rv.CardActions():
            with rv.ListItem(class_="grow"):
                with rv.ListItemAvatar(color="grey darken-3"):
                    rv.Img(
                        class_="elevation-6",
                        src=img,
                    )
                with rv.ListItemContent():
                    rv.ListItemTitle(children=[name])
                    rv.ListItemSubtitle(children=[position], style_="white-space: unset;")
        rv.CardText(
            children=[text],
            style_="font-style: italic; padding-top: 0px",
        )



import logging
import pathlib
import shutil

import requests
from solara.server.utils import path_is_child_of

import solara.settings

logger = logging.getLogger("Solara.cdn")

cdn_url_path = "_solara/cdn"


def put_in_cache(base_cache_dir: pathlib.Path, path, data: bytes):
    cache_path = base_cache_dir / path
    if not path_is_child_of(cache_path, base_cache_dir):
        raise PermissionError("Trying to write outside of cache directory")

    pathlib.Path(cache_path.parent).mkdir(parents=True, exist_ok=True)
    try:
        logger.info("Writing cache file: %s", cache_path)
        cache_path.write_bytes(data)
    except FileNotFoundError:
        logger.info("Failed writing cache file: %s", cache_path)


def get_from_cache(base_cache_dir: pathlib.Path, path):
    cache_path = pathlib.Path(base_cache_dir / path)
    # Make sure cache_path is a subdirectory of base_cache_dir
    # so we don't accidentally read files from the parent directory
    # which is a security risk.
    if not path_is_child_of(cache_path, base_cache_dir):
        logger.warning("Trying to read from outside of cache directory: %s is not a subdir of %s", cache_path, base_cache_dir)
        raise PermissionError("Trying to read from outside of cache directory")

    try:
        logger.info("Opening cache file: %s", cache_path)
        return cache_path.read_bytes()
    except FileNotFoundError:
        pass


def get_cdn_url(path):
    path = str(path)  # on windows, the path can contain a \
    return str(solara.settings.assets.cdn) + str(path).replace("\\", "/")


def get_data(base_cache_dir: pathlib.Path, path):
    parts = path.replace("\\", "/").split("/")
    store_path = path if len(parts) != 1 else pathlib.Path(path) / "__main.js"

    content = get_from_cache(base_cache_dir, store_path)
    if content:
        return content

    url = get_cdn_url(path)
    response = requests.get(url)
    if response.ok:
        put_in_cache(base_cache_dir, store_path, response.content)
        return response.content
    else:
        logger.warning("Could not load URL: %r", url)
        raise Exception(f"Could not load URL: {url}")


def get_path(base_cache_dir: pathlib.Path, path) -> pathlib.Path:
    parts = path.replace("\\", "/").split("/")
    store_path = path if len(parts) != 1 else pathlib.Path(path) / "__main.js"
    cache_path = base_cache_dir / store_path

    if not path_is_child_of(cache_path, base_cache_dir):
        raise PermissionError("Trying to read from outside of cache directory")

    if cache_path.exists():
        # before d7eba856f100d5c3c64f4eec22c62390f084cb40 on windows, we could
        # accidentally write to the cache directory, so we need to check if we still
        # have an old directory layout, and remove that first.
        if cache_path.is_dir():
            shutil.rmtree(cache_path)
        else:
            return cache_path
    url = get_cdn_url(path)
    response = requests.get(url)
    if response.ok:
        put_in_cache(base_cache_dir, store_path, response.content)
        assert cache_path.exists(), f"Could not write to {cache_path}"
        return cache_path
    else:
        logger.warning("Could not load URL: %r", url)
        raise Exception(f"Could not load URL: {url}")



"""# Scatter plot using Plotly

This example shows how to use Plotly to create a scatter plot and a slider to do some filtering.

Inspired by the dash documentation.


## Note

Solara supports plotly and plotly express. Create your figure (not a figure widget)
and pass it to the [FigurePlotly](/documentation/components/viz/plotly) component.

"""

import pandas as pd
import plotly.express as px
import solara

title = "Scatter plot using Plotly"

try:
    df = pd.read_csv("https://raw.githubusercontent.com/plotly/datasets/master/gapminderDataFiveYear.csv")
except Exception:
    df = None

if df is not None:
    year_min = df["year"].min()
    year_max = df["year"].max()
    years = df["year"].unique().tolist()


@solara.component
def Page():
    with solara.Div() as main:
        index = solara.ui_slider(value=0, min=0, max=len(years) - 1, tick_labels=years, key="year slider index")
        selected_year = years[index]

        filtered_df = df[df.year == selected_year].copy()

        fig = px.scatter(filtered_df, x="gdpPercap", y="lifeExp", size="pop", color="continent", hover_name="country", log_x=True, size_max=55)
        fig.update_layout(transition_duration=1500)

        solara.FigurePlotly(fig, dependencies=[selected_year])
    return main



from pathlib import Path

import playwright.sync_api
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Mount, Route

import solara
import solara.server.starlette
from solara.server import settings

try:
    from . import conftest
except ImportError:
    pass


HERE = Path(__file__).parent


@solara.component
def Page():
    solara.Markdown("Mounted in starlette")


def myroot(request: Request):
    return JSONResponse({"framework": "solara"})


starlette_routes = [
    Route("/", endpoint=myroot),
    Mount("/solara_mount/", routes=solara.server.starlette.routes),
]

starlette_app = Starlette(routes=starlette_routes)


def test_starlette_mount(page_session: playwright.sync_api.Page, solara_app, extra_include_path):
    settings.main.root_path = None
    settings.main.base_url = ""
    try:
        port = conftest.TEST_PORT
        conftest.TEST_PORT += 1
        server = solara.server.starlette.ServerStarlette(port=port, starlette_app=starlette_app)
        server.serve_threaded()
        server.wait_until_serving()
        with extra_include_path(HERE), solara_app("starlette_test", init=False):
            page_session.goto(f"{server.base_url}/solara_mount/")
            page_session.locator("text=Mounted in starlette").wait_for()
    finally:
        settings.main.root_path = None
        settings.main.base_url = ""
